from typing import Type, Optional, Any, Dict, Generic, Tuple

from pydantic import create_model
from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.collections.base import _CollectionBase, _CollectionsBase
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.orm import (
    BaseProperty,
    CollectionModelConfig,
    Model,
    UserModelType,
)
from weaviate.collections.config import _ConfigCollectionModel
from weaviate.collections.data import _DataCollectionModel
from weaviate.collections.query import _GrpcCollectionModel
from weaviate.collections.tenants import _Tenants
from weaviate.connect import Connection
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.util import _capitalize_first_letter
from weaviate.types import PYTHON_TYPE_TO_DATATYPE


class _CollectionObjectModel(_CollectionBase, Generic[Model]):
    def __init__(
        self,
        connection: Connection,
        name: str,
        model: Type[Model],
        config: _ConfigCollectionModel,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
    ) -> None:
        super().__init__(name)

        self._connection = connection

        self.config = config
        self.data = _DataCollectionModel[Model](
            connection, self.name, model, consistency_level, tenant
        )
        self.query = _GrpcCollectionModel[Model](
            connection, self.name, model, tenant, consistency_level
        )
        self.tenants = _Tenants(connection, self.name)

        self.__consistency_level = consistency_level
        self.__model: Type[Model] = model
        self.__tenant = tenant

    @property
    def model(self) -> Type[Model]:
        return self.__model

    def with_tenant(self, tenant: Optional[str] = None) -> "_CollectionObjectModel[Model]":
        return _CollectionObjectModel[Model](
            self._connection, self.name, self.__model, self.config, self.__consistency_level, tenant
        )

    def with_consistency_level(
        self, consistency_level: Optional[ConsistencyLevel] = None
    ) -> "_CollectionObjectModel[Model]":
        return _CollectionObjectModel[Model](
            self._connection, self.name, self.__model, self.config, consistency_level, self.__tenant
        )


class _CollectionModel(_CollectionsBase):
    def create(self, config: CollectionModelConfig[Model]) -> _CollectionObjectModel[Model]:
        name = super()._create(config)
        config_name = _capitalize_first_letter(config.model.__name__)
        if config_name != name:
            raise ValueError(
                f"Name of created collection ({name}) does not match given name ({config_name})"
            )
        return self.get(config.model)

    def get(self, model: Type[Model]) -> _CollectionObjectModel[Model]:
        name = _capitalize_first_letter(model.__name__)
        config = _ConfigCollectionModel(self._connection, name, None)
        if config.is_invalid(model):
            raise TypeError(
                f"Model {model.__name__} definition does not match collection {name} config"
            )
        return _CollectionObjectModel[Model](self._connection, name, model, config)

    def get_dynamic(self, name: str) -> Tuple[_CollectionObjectModel[BaseProperty], UserModelType]:
        path = f"/schema/{_capitalize_first_letter(name)}"

        try:
            response = self._connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Schema could not be retrieved.") from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Get schema", response)

        response_json = response.json()
        fields: Dict[str, Any] = {
            prop["name"]: (PYTHON_TYPE_TO_DATATYPE[prop["dataType"][0]], None)
            for prop in response_json["properties"]
        }
        model = create_model(response_json["class"], **fields, __base__=BaseProperty)
        config = _ConfigCollectionModel(self._connection, name, None)
        return _CollectionObjectModel[BaseProperty](self._connection, name, model, config), model

    def delete(self, model: Type[Model]) -> None:
        """Use this method to delete a collection from the Weaviate instance by its ORM model.

        WARNING: If you have instances of client.orm.get() or client.orm.create()
        for this collection within your code, they will cease to function correctly after this operation.

        Parameters:
            - model: The ORM model of the collection to be deleted.
        """
        name = _capitalize_first_letter(model.__name__)
        return self._delete(name)

    def exists(self, model: Type[Model]) -> bool:
        name = _capitalize_first_letter(model.__name__)
        return self._exists(name)

    def update(self, model: Type[Model]) -> _CollectionObjectModel[Model]:
        name = _capitalize_first_letter(model.__name__)
        config = _ConfigCollectionModel(self._connection, name, None)
        config.update_model(model)
        return _CollectionObjectModel[Model](self._connection, name, model, config)



from typing import List, Optional, Union

from pydantic import Field

from weaviate.collections.classes.types import _WeaviateInput
from weaviate.util import BaseEnum
from weaviate.types import UUID


class HybridFusion(str, BaseEnum):
    """Define how the query's hybrid fusion operation should be performed."""

    RANKED = "FUSION_TYPE_RANKED"
    RELATIVE_SCORE = "FUSION_TYPE_RELATIVE_SCORE"


class Move:
    """Define how the query's move operation should be performed."""

    def __init__(
        self,
        force: float,
        objects: Optional[Union[List[UUID], UUID]] = None,
        concepts: Optional[Union[List[str], str]] = None,
    ):
        if (objects is None or (isinstance(objects, list) and len(objects) == 0)) and (
            concepts is None or (isinstance(concepts, list) and len(concepts) == 0)
        ):
            raise ValueError("Either objects or concepts need to be given")

        self.force = force

        # accept single values, but make them a list
        if objects is None:
            self.__objects = None
        elif not isinstance(objects, list):
            self.__objects = [str(objects)]
        else:
            self.__objects = [str(obj_uuid) for obj_uuid in objects]

        if concepts is None:
            self.__concepts = None
        elif not isinstance(concepts, list):
            self.__concepts = [concepts]
        else:
            self.__concepts = concepts

    @property
    def _objects_list(self) -> Optional[List[str]]:
        return self.__objects

    @property
    def _concepts_list(self) -> Optional[List[str]]:
        return self.__concepts

    def _to_gql_payload(self) -> dict:
        payload: dict = {"force": self.force}
        if self.__objects is not None:
            payload["objects"] = self.__objects
        if self.__concepts is not None:
            payload["concepts"] = self.__concepts
        return payload


class MetadataQuery(_WeaviateInput):
    """Define which metadata should be returned in the query's results."""

    uuid: bool = Field(default=False)
    vector: bool = Field(default=False)
    creation_time_unix: bool = Field(default=False)
    last_update_time_unix: bool = Field(default=False)
    distance: bool = Field(default=False)
    certainty: bool = Field(default=False)
    score: bool = Field(default=False)
    explain_score: bool = Field(default=False)
    is_consistent: bool = Field(default=False)

    @classmethod
    def _full(cls) -> "MetadataQuery":
        """Return a MetadataQuery with all fields set to True except for vector."""
        return cls(
            uuid=True,
            creation_time_unix=True,
            last_update_time_unix=True,
            distance=True,
            certainty=True,
            score=True,
            explain_score=True,
            is_consistent=True,
        )


class Generate(_WeaviateInput):
    """Define how the query's RAG capabilities should be performed."""

    single_prompt: Optional[str] = Field(default=None)
    grouped_task: Optional[str] = Field(default=None)
    grouped_properties: Optional[List[str]] = Field(default=None)


class Sort(_WeaviateInput):
    """Define how the query's sort operation should be performed."""

    prop: str
    ascending: bool = Field(default=True)


class FromReference(_WeaviateInput):
    """Define a query-time reference to a single-target property when querying through cross-references."""

    link_on: str
    return_properties: Optional["PROPERTIES"] = Field(default=None)
    return_metadata: Optional[MetadataQuery] = Field(default=None)

    def __hash__(self) -> int:  # for set
        return hash(str(self))


class FromReferenceMultiTarget(FromReference):
    """Define a query-time reference to a multi-target property when querying through cross-references."""

    target_collection: str


class FromNested(_WeaviateInput):
    """Define the return properties of a nested property."""

    name: str
    properties: "NestedProperties"

    def __hash__(self) -> int:  # for set
        return hash(str(self))


PROPERTY = Union[str, FromReference, FromNested]
PROPERTIES = Union[List[PROPERTY], PROPERTY]

NestedProperties = Union[List[Union[str, FromNested]], str, FromNested]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/batch.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from weaviate.proto.v1 import base_pb2 as v1_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0ev1/batch.proto\x12\x0bweaviate.v1\x1a\x1cgoogle/protobuf/struct.proto\x1a\rv1/base.proto"\x95\x01\n\x13\x42\x61tchObjectsRequest\x12)\n\x07objects\x18\x01 \x03(\x0b\x32\x18.weaviate.v1.BatchObject\x12=\n\x11\x63onsistency_level\x18\x02 \x01(\x0e\x32\x1d.weaviate.v1.ConsistencyLevelH\x00\x88\x01\x01\x42\x14\n\x12_consistency_level"\x83\x07\n\x0b\x42\x61tchObject\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0e\n\x06vector\x18\x02 \x03(\x02\x12\x37\n\nproperties\x18\x03 \x01(\x0b\x32#.weaviate.v1.BatchObject.Properties\x12\x12\n\ncollection\x18\x04 \x01(\t\x12\x0e\n\x06tenant\x18\x05 \x01(\t\x1a\xea\x04\n\nProperties\x12\x33\n\x12non_ref_properties\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12N\n\x17single_target_ref_props\x18\x02 \x03(\x0b\x32-.weaviate.v1.BatchObject.SingleTargetRefProps\x12L\n\x16multi_target_ref_props\x18\x03 \x03(\x0b\x32,.weaviate.v1.BatchObject.MultiTargetRefProps\x12\x43\n\x17number_array_properties\x18\x04 \x03(\x0b\x32".weaviate.v1.NumberArrayProperties\x12=\n\x14int_array_properties\x18\x05 \x03(\x0b\x32\x1f.weaviate.v1.IntArrayProperties\x12?\n\x15text_array_properties\x18\x06 \x03(\x0b\x32 .weaviate.v1.TextArrayProperties\x12\x45\n\x18\x62oolean_array_properties\x18\x07 \x03(\x0b\x32#.weaviate.v1.BooleanArrayProperties\x12\x38\n\x11object_properties\x18\x08 \x03(\x0b\x32\x1d.weaviate.v1.ObjectProperties\x12\x43\n\x17object_array_properties\x18\t \x03(\x0b\x32".weaviate.v1.ObjectArrayProperties\x1a\x38\n\x14SingleTargetRefProps\x12\r\n\x05uuids\x18\x01 \x03(\t\x12\x11\n\tprop_name\x18\x02 \x01(\t\x1aR\n\x13MultiTargetRefProps\x12\r\n\x05uuids\x18\x01 \x03(\t\x12\x11\n\tprop_name\x18\x02 \x01(\t\x12\x19\n\x11target_collection\x18\x03 \x01(\t"\x88\x01\n\x11\x42\x61tchObjectsReply\x12\x0c\n\x04took\x18\x01 \x01(\x02\x12\x39\n\x06\x65rrors\x18\x02 \x03(\x0b\x32).weaviate.v1.BatchObjectsReply.BatchError\x1a*\n\nBatchError\x12\r\n\x05index\x18\x01 \x01(\x05\x12\r\n\x05\x65rror\x18\x02 \x01(\tBo\n#io.weaviate.client.grpc.protocol.v1B\x12WeaviateProtoBatchZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "v1.batch_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n#io.weaviate.client.grpc.protocol.v1B\022WeaviateProtoBatchZ4github.com/weaviate/weaviate/grpc/generated;protocol"
    _globals["_BATCHOBJECTSREQUEST"]._serialized_start = 77
    _globals["_BATCHOBJECTSREQUEST"]._serialized_end = 226
    _globals["_BATCHOBJECT"]._serialized_start = 229
    _globals["_BATCHOBJECT"]._serialized_end = 1128
    _globals["_BATCHOBJECT_PROPERTIES"]._serialized_start = 368
    _globals["_BATCHOBJECT_PROPERTIES"]._serialized_end = 986
    _globals["_BATCHOBJECT_SINGLETARGETREFPROPS"]._serialized_start = 988
    _globals["_BATCHOBJECT_SINGLETARGETREFPROPS"]._serialized_end = 1044
    _globals["_BATCHOBJECT_MULTITARGETREFPROPS"]._serialized_start = 1046
    _globals["_BATCHOBJECT_MULTITARGETREFPROPS"]._serialized_end = 1128
    _globals["_BATCHOBJECTSREPLY"]._serialized_start = 1131
    _globals["_BATCHOBJECTSREPLY"]._serialized_end = 1267
    _globals["_BATCHOBJECTSREPLY_BATCHERROR"]._serialized_start = 1225
    _globals["_BATCHOBJECTSREPLY_BATCHERROR"]._serialized_end = 1267
# @@protoc_insertion_point(module_scope)



import pytest

import weaviate
from mock_tests.conftest import MOCK_SERVER_URL
from weaviate.exceptions import ResponseCannotBeDecodedException


def test_json_decode_exception_dict(weaviate_mock):
    """Tests that JsonDecodeException is raised containing the correct error."""

    weaviate_mock.expect_request("/v1/schema").respond_with_data("JsonCannotParseThis")

    client = weaviate.Client(MOCK_SERVER_URL)
    with pytest.raises(ResponseCannotBeDecodedException) as e:
        client.schema.get()

        assert "JsonCannotParseThis" in e.value


def test_json_decode_exception_list(weaviate_mock):
    """Tests that JsonDecodeException is raised containing the correct error."""

    weaviate_mock.expect_request("/v1/schema/Test/shards").respond_with_data("JsonCannotParseThis")

    client = weaviate.Client(MOCK_SERVER_URL)
    with pytest.raises(ResponseCannotBeDecodedException) as e:
        client.schema.get_class_shards("Test")
        assert "JsonCannotParseThis" in e.value



from weaviate.collections.aggregations.over_all import _OverAll, _OverAllGroupBy
from weaviate.collections.aggregations.near_image import _NearImage, _NearImageGroupBy
from weaviate.collections.aggregations.near_object import _NearObject, _NearObjectGroupBy
from weaviate.collections.aggregations.near_text import _NearText, _NearTextGroupBy
from weaviate.collections.aggregations.near_vector import _NearVector, _NearVectorGroupBy


class _AggregateCollection(_OverAll, _NearImage, _NearObject, _NearText, _NearVector):
    pass


class _AggregateGroupByCollection(
    _OverAllGroupBy, _NearImageGroupBy, _NearObjectGroupBy, _NearTextGroupBy, _NearVectorGroupBy
):
    pass



import unittest
from unittest.mock import Mock

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.gql.query import Query


class TestQuery(unittest.TestCase):
    def test_get(self):
        """
        Test the `get` attribute.
        """

        query = Query(Mock())

        gql = query.get("Group", ["name", "uuid"]).build()
        self.assertEqual("{Get{Group{name uuid}}}", gql)

    def test_aggregate(self):
        """
        Test the `aggregate` attribute.
        """

        query = Query(Mock())

        gql = query.aggregate("Group").build()
        self.assertEqual("{Aggregate{Group{}}}", gql)

    def test_raw(self):
        """
        Test the `raw` method.
        """

        # valid calls
        connection_mock = mock_connection_func("post", return_json={})
        query = Query(connection_mock)

        gql_query = "{Get {Group {name Members {... on Person {name}}}}}"
        query.raw(gql_query)

        connection_mock.post.assert_called_with(
            path="/graphql", weaviate_object={"query": gql_query}
        )

        # invalid calls

        type_error_message = "Query is expected to be a string"
        requests_error_message = "Query not executed."
        query_error_message = "GQL query failed"

        with self.assertRaises(TypeError) as error:
            query.raw(["TestQuery"])
        check_error_message(self, error, type_error_message)

        query = Query(mock_connection_func("post", side_effect=RequestsConnectionError("Test!")))
        with self.assertRaises(RequestsConnectionError) as error:
            query.raw("TestQuery")
        check_error_message(self, error, requests_error_message)

        query = Query(mock_connection_func("post", status_code=404))
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            query.raw("TestQuery")
        check_startswith_error_message(self, error, query_error_message)



"""
BatchRequest class definitions.
"""
import copy
from abc import ABC, abstractmethod
from typing import List, Sequence, Optional, Dict, Any, Union
from uuid import uuid4

from weaviate.util import get_valid_uuid, get_vector
from weaviate.types import UUID

BatchResponse = List[Dict[str, Any]]


class BatchRequest(ABC):
    """
    BatchRequest abstract class used as a interface for batch requests.
    """

    def __init__(self) -> None:
        self._items: List[Dict[str, Any]] = []

    def __len__(self) -> int:
        return len(self._items)

    def is_empty(self) -> bool:
        """
        Check if BatchRequest is empty.

        Returns
        -------
        bool
            Whether the BatchRequest is empty.
        """

        return len(self._items) == 0

    def empty(self) -> None:
        """
        Remove all the items from the BatchRequest.
        """

        self._items = []

    def pop(self, index: int = -1) -> dict:
        """
        Remove and return item at index (default last).

        Parameters
        ----------
        index : int, optional
            The index of the item to pop, by default -1 (last item).

        Returns
        -------
        dict
            The popped item.

        Raises
        -------
        IndexError
            If batch is empty or index is out of range.
        """

        return self._items.pop(index)

    @abstractmethod
    def add(self, *args, **kwargs):  # type: ignore
        """Add objects to BatchRequest."""

    @abstractmethod
    def get_request_body(self) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """Return the request body to be digested by weaviate that contains all batch items."""

    @abstractmethod
    def add_failed_objects_from_response(
        self,
        response_item: BatchResponse,
        errors_to_exclude: Optional[List[str]],
        errors_to_include: Optional[List[str]],
    ) -> BatchResponse:
        """Add failed items from a weaviate response.

        Parameters
        ----------
        response_item : BatchResponse
            Weaviate response that contains the status for all objects.
        errors_to_exclude : Optional[List[str]]
            Which errors should NOT be retried.
        errors_to_include : Optional[List[str]]
            Which errors should be retried.

        Returns
        ------
        BatchResponse: Contains responses form all successful object, eg. those that have not been added to this batch.
        """

    @staticmethod
    def _skip_objects_retry(
        entry: Dict[str, Any],
        errors_to_exclude: Optional[List[str]],
        errors_to_include: Optional[List[str]],
    ) -> bool:
        if (
            len(entry["result"]) == 0
            or "errors" not in entry["result"]
            or "error" not in entry["result"]["errors"]
            or len(entry["result"]["errors"]["error"]) == 0
        ):
            return True

        # skip based on error messages
        if errors_to_exclude is not None:
            for err in entry["result"]["errors"]["error"]:
                if any(excl in err["message"] for excl in errors_to_exclude):
                    return True
            return False
        elif errors_to_include is not None:
            for err in entry["result"]["errors"]["error"]:
                if any(incl in err["message"] for incl in errors_to_include):
                    return False
            return True
        return False


class ReferenceBatchRequest(BatchRequest):
    """
    Collect Weaviate-object references to add them in one request to Weaviate.
    Caution this request will miss some validations to be faster.
    """

    def add(
        self,
        from_object_class_name: str,
        from_object_uuid: UUID,
        from_property_name: str,
        to_object_uuid: UUID,
        to_object_class_name: Optional[str] = None,
        tenant: Optional[str] = None,
    ) -> None:
        """
        Add one Weaviate-object reference to this batch. Does NOT validate the consistency of the
        reference against the class schema. Checks the arguments' type and UUIDs' format.

        Parameters
        ----------
        from_object_class_name : str
            The name of the class that should reference another object.
        from_object_uuid : str
            The UUID or URL of the object that should reference another object.
        from_property_name : str
            The name of the property that contains the reference.
        to_object_uuid : str
            The UUID or URL of the object that is actually referenced.
        to_object_class_name : Optional[str], optional
            The referenced object class name to which to add the reference (with UUID
            `to_object_uuid`), it is included in Weaviate 1.14.0, where all objects are namespaced
            by class name.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None

        Raises
        ------
        TypeError
            If arguments are not of type str.
        ValueError
            If 'uuid' is not valid or cannot be extracted.
        """

        if not isinstance(from_object_class_name, str):
            raise TypeError("'from_object_class_name' argument must be of type str")

        if not isinstance(from_property_name, str):
            raise TypeError("'from_property_name' argument must be of type str")

        if to_object_class_name is not None and not isinstance(to_object_class_name, str):
            raise TypeError("'to_object_class_name' argument must be of type str")

        to_object_uuid = get_valid_uuid(to_object_uuid)
        from_object_uuid = get_valid_uuid(from_object_uuid)

        if to_object_class_name is not None:
            to_beacon = f"weaviate://localhost/{to_object_class_name}/{to_object_uuid}"
        else:
            to_beacon = f"weaviate://localhost/{to_object_uuid}"

        item = {
            "from": "weaviate://localhost/"
            + from_object_class_name
            + "/"
            + from_object_uuid
            + "/"
            + from_property_name,
            "to": to_beacon,
        }

        if tenant is not None:
            item["tenant"] = tenant

        self._items.append(item)

    def get_request_body(self) -> List[Dict[str, Any]]:
        """
        Get request body as a list of dictionaries, where each dictionary
        is a Weaviate-object reference.

        Returns
        -------
        List[dict]
            A list of Weaviate-objects references as dictionaries.
        """

        return self._items

    def add_failed_objects_from_response(
        self,
        response: BatchResponse,
        errors_to_exclude: Optional[List[str]],
        errors_to_include: Optional[List[str]],
    ) -> BatchResponse:
        successful_responses = []

        for ref in response:
            if self._skip_objects_retry(ref, errors_to_exclude, errors_to_include):
                successful_responses.append(ref)
                continue
            self._items.append({"from": ref["from"], "to": ref["to"]})
        return successful_responses


class ObjectsBatchRequest(BatchRequest):
    """
    Collect objects for one batch request to weaviate.
    Caution this batch will not be validated through weaviate.
    """

    def add(
        self,
        data_object: dict,
        class_name: str,
        uuid: Optional[UUID] = None,
        vector: Optional[Sequence] = None,
        tenant: Optional[str] = None,
    ) -> str:
        """
        Add one object to this batch. Does NOT validate the consistency of the object against
        the client's schema. Checks the arguments' type and UUIDs' format.

        Parameters
        ----------
        class_name : str
            The name of the class this object belongs to.
        data_object : dict
            Object to be added as a dict datatype.
        uuid : str or None, optional
            UUID of the object as a string, by default None
        vector: Sequence or None, optional
            The embedding of the object that should be validated.
            Can be used when:
             - a class does not have a vectorization module.
             - The given vector was generated using the _identical_ vectorization module that is configured for the
             class. In this case this vector takes precedence.

            Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` and `tf.Tensor`,
            by default None.
        tenant: str, optional
            Tenant of the object

        Returns
        -------
        str
            The UUID of the added object. If one was not provided a UUIDv3 will be generated.

        Raises
        ------
        TypeError
            If an argument passed is not of an appropriate type.
        ValueError
            If 'uuid' is not of a proper form.
        """

        if not isinstance(data_object, dict):
            raise TypeError("Object must be of type dict")
        if not isinstance(class_name, str):
            raise TypeError("Class name must be of type str")

        batch_item = {"class": class_name, "properties": copy.deepcopy(data_object)}
        if uuid is not None:
            valid_uuid = get_valid_uuid(uuid)
        else:
            valid_uuid = get_valid_uuid(uuid4())
        batch_item["id"] = valid_uuid

        if vector is not None:
            batch_item["vector"] = get_vector(vector)
        if tenant is not None:
            batch_item["tenant"] = tenant

        self._items.append(batch_item)

        return valid_uuid

    def get_request_body(self) -> Dict[str, Any]:
        """
        Get the request body as it is needed for the Weaviate server.

        Returns
        -------
        dict
            The request body as a dict.
        """

        return {"fields": ["ALL"], "objects": self._items}

    def add_failed_objects_from_response(
        self,
        response: BatchResponse,
        errors_to_exclude: Optional[List[str]],
        errors_to_include: Optional[List[str]],
    ) -> BatchResponse:
        successful_responses = []

        for obj in response:
            if self._skip_objects_retry(obj, errors_to_exclude, errors_to_include):
                successful_responses.append(obj)
                continue
            self.add(
                data_object=obj["properties"],
                class_name=obj["class"],
                uuid=obj["id"],
                vector=obj.get("vector", None),
                tenant=obj.get("tenant", None),
            )
        return successful_responses



from typing import Optional, Tuple

from weaviate.auth import AuthCredentials
from weaviate.client import WeaviateClient
from weaviate.config import AdditionalConfig
from weaviate.connect.connection import ConnectionParams, ProtocolParams
from weaviate.embedded import EmbeddedOptions


def connect_to_wcs(
    cluster_id: str,
    auth_credentials: Optional[AuthCredentials],
    headers: Optional[dict] = None,
    timeout: Tuple[int, int] = (10, 60),
) -> WeaviateClient:
    """
    Connect to your own Weaviate Cloud Service (WCS) instance.

    Arguments:
        `cluster_id`
            The WCS cluster id to connect to.
        `auth_credentials`
            The credentials to use for authentication with your WCS instance. This can be an API key, in which case use `weaviate.auth.AuthApiKey`,
            a bearer token, in which case use `weaviate.auth.AuthBearerToken`, a client secret, in which case use `weaviate.auth.AuthClientCredentials`
            or a username and password, in which case use `weaviate.auth.AuthClientPassword`.
        `headers`
            Additional headers to include in the requests, e.g. API keys for third-party Cloud vectorisation.
        `timeout`
            The timeout to use for the underlying HTTP calls. Accepts a tuple of integers, where the first integer
            represents the connect timeout and the second integer represents the read timeout.

    Returns
        `weaviate.WeaviateClient`
            The client connected to the cluster with the required parameters set appropriately.
    """
    return WeaviateClient(
        connection_params=ConnectionParams(
            http=ProtocolParams(host=f"{cluster_id}.weaviate.network", port=443, secure=True),
            grpc=ProtocolParams(host=f"{cluster_id}.weaviate.network", port=50051, secure=True),
        ),
        auth_client_secret=auth_credentials,
        additional_headers=headers,
        additional_config=AdditionalConfig(timeout=timeout),
    )


def connect_to_local(
    host: str = "localhost",
    port: int = 8080,
    grpc_port: int = 50051,
    headers: Optional[dict] = None,
    timeout: Tuple[int, int] = (10, 60),
) -> WeaviateClient:
    """
    Connect to a local Weaviate instance deployed using Docker compose with standard port configurations.

    Arguments:
        `schema`
            The schema to use for the underlying REST & GraphQL API calls.
        `host`
            The host to use for the underlying REST & GraphQL API calls.
        `port`
            The port to use for the underlying REST & GraphQL API calls.
        `grpc_port`
            The port to use for the underlying gRPC API.
        `headers`
            Additional headers to include in the requests, e.g. API keys for Cloud vectorisation.
        `timeout`
            The timeout to use for the underlying HTTP calls. Accepts a tuple of integers, where the first integer
            represents the connect timeout and the second integer represents the read timeout.

    Returns
        `weaviate.WeaviateClient`
            The client connected to the local instance with default parameters set as:
    """
    return WeaviateClient(
        connection_params=ConnectionParams(
            http=ProtocolParams(host=host, port=port, secure=False),
            grpc=ProtocolParams(host=host, port=grpc_port, secure=False),
        ),
        additional_headers=headers,
        additional_config=AdditionalConfig(timeout=timeout),
    )


def connect_to_embedded(
    port: int = 8079,
    grpc_port: int = 50051,
    headers: Optional[dict] = None,
    timeout: Tuple[int, int] = (10, 60),
) -> WeaviateClient:
    """
    Connect to an embedded Weaviate instance.

    Arguments:
        `port`
            The port to use for the underlying REST & GraphQL API calls.
        `grpc_port`
            The port to use for the underlying gRPC API.
        `headers`
            Additional headers to include in the requests, e.g. API keys for Cloud vectorisation.
        `timeout`
            The timeout to use for the underlying HTTP calls. Accepts a tuple of integers, where the first integer
            represents the connect timeout and the second integer represents the read timeout.

    Returns
        `weaviate.WeaviateClient`
            The client connected to the embedded instance with the required parameters set appropriately.
    """
    return WeaviateClient(
        embedded_options=EmbeddedOptions(
            port=port,
            grpc_port=grpc_port,
        ),
        additional_headers=headers,
        additional_config=AdditionalConfig(timeout=timeout),
    )


def connect_to_custom(
    http_host: str,
    http_port: int,
    http_secure: bool,
    grpc_host: str,
    grpc_port: int,
    grpc_secure: bool,
    headers: Optional[dict] = None,
    timeout: Tuple[int, int] = (10, 60),
) -> WeaviateClient:
    """
    Connect to a Weaviate instance with custom connection parameters.

    If this is not sufficient for your customisation needs then instantiate a `weaviate.WeaviateClient` directly.

    Arguments:
        `http_host`
            The host to use for the underlying REST & GraphQL API calls.
        `http_port`
            The port to use for the underlying REST & GraphQL API calls.
        `http_secure`
            Whether to use https for the underlying REST & GraphQL API calls.
        `grpc_host`
            The host to use for the underlying gRPC API.
        `grpc_port`
            The port to use for the underlying gRPC API.
        `grpc_secure`
            Whether to use a secure channel for the underlying gRPC API.
        `headers`
            Additional headers to include in the requests, e.g. API keys for Cloud vectorisation.
        `timeout`
            The timeout to use for the underlying HTTP calls. Accepts a tuple of integers, where the first integer
            represents the connect timeout and the second integer represents the read timeout.

    Returns
        `weaviate.WeaviateClient`
            The client connected to the instance with the required parameters set appropriately.
    """
    return WeaviateClient(
        ConnectionParams.from_params(
            http_host=http_host,
            http_port=http_port,
            http_secure=http_secure,
            grpc_host=grpc_host,
            grpc_port=grpc_port,
            grpc_secure=grpc_secure,
        ),
        additional_headers=headers,
        additional_config=AdditionalConfig(timeout=timeout),
    )



import os
import time
import warnings
from typing import Optional, Dict

import pytest
import requests
from requests.exceptions import ConnectionError as RequestsConnectionError
from requests.exceptions import HTTPError as RequestsHTTPError

import weaviate
from weaviate import (
    AuthenticationFailedException,
    AuthClientCredentials,
    AuthClientPassword,
    AuthBearerToken,
)
from weaviate.auth import AuthApiKey
from weaviate.exceptions import WeaviateStartUpError, UnexpectedStatusCodeException

ANON_PORT = 8080
AZURE_PORT = 8081
OKTA_PORT_CC = 8082
OKTA_PORT_USERS = 8083
WCS_PORT = 8085


def wait_for_weaviate(url: str):
    ready_url = url + "/v1/.well-known/ready"
    for _i in range(5):
        try:
            requests.get(ready_url).raise_for_status()
            return
        except (RequestsHTTPError, RequestsConnectionError):
            time.sleep(1)

    try:
        requests.get(ready_url).raise_for_status()
        return
    except (RequestsHTTPError, RequestsConnectionError):
        pass
    raise WeaviateStartUpError(
        f"Weaviate did not start up in 5 seconds. Either the Weaviate URL {url} is wrong or Weaviate did not start up in the interval given in 'startup_period'."
    )


def is_auth_enabled(url: str):
    wait_for_weaviate(url)
    response = requests.get(url + "/v1/.well-known/openid-configuration")
    return response.status_code == 200


def test_no_auth_provided():
    """Test exception when trying to access a weaviate that requires authentication."""
    url = f"http://localhost:{AZURE_PORT}"
    assert is_auth_enabled(url)
    with pytest.raises(AuthenticationFailedException):
        weaviate.Client(url)


@pytest.mark.parametrize(
    "name,env_variable_name,port,scope",
    [
        ("azure", "AZURE_CLIENT_SECRET", AZURE_PORT, None),
        (
            "azure",
            "AZURE_CLIENT_SECRET",
            AZURE_PORT,
            "4706508f-30c2-469b-8b12-ad272b3de864/.default",
        ),
        ("okta", "OKTA_CLIENT_SECRET", OKTA_PORT_CC, "some_scope"),
    ],
)
def test_authentication_client_credentials(
    name: str, env_variable_name: str, port: str, scope: Optional[str]
):
    """Test client credential flow with various providers."""
    client_secret = os.environ.get(env_variable_name)
    if client_secret is None:
        pytest.skip(f"No {name} login data found.")

    url = f"http://localhost:{port}"
    assert is_auth_enabled(url)
    client = weaviate.Client(
        url,
        auth_client_secret=AuthClientCredentials(client_secret=client_secret, scope=scope),
    )
    client.schema.delete_all()  # no exception


@pytest.mark.parametrize(
    "name,user,env_variable_name,port,scope,warning",
    [
        (
            "WCS",
            "ms_2d0e007e7136de11d5f29fce7a53dae219a51458@existiert.net",
            "WCS_DUMMY_CI_PW",
            WCS_PORT,
            None,
            False,
        ),
        (
            "okta",
            "test@test.de",
            "OKTA_DUMMY_CI_PW",
            OKTA_PORT_USERS,
            "some_scope offline_access",
            False,
        ),
        (
            "okta - no refresh",
            "test@test.de",
            "OKTA_DUMMY_CI_PW",
            OKTA_PORT_USERS,
            "some_scope",
            True,
        ),
    ],
)
def test_authentication_user_pw(
    recwarn, name: str, user: str, env_variable_name: str, port: str, scope: str, warning: bool
):
    """Test authentication using Resource Owner Password Credentials Grant (User + PW)."""
    # testing for warnings can be flaky without this as there are open SSL conections
    warnings.filterwarnings(action="ignore", message="unclosed", category=ResourceWarning)

    url = f"http://localhost:{port}"
    assert is_auth_enabled(url)

    pw = os.environ.get(env_variable_name)
    if pw is None:
        pytest.skip(f"No login data for {name} found.")

    if scope is not None:
        auth = AuthClientPassword(username=user, password=pw, scope=scope)
    else:
        auth = AuthClientPassword(username=user, password=pw)

    client = weaviate.Client(url, auth_client_secret=auth)
    client.schema.delete_all()  # no exception
    if warning:
        assert len(recwarn) == 1
        w = recwarn.pop()
        assert issubclass(w.category, UserWarning)
        assert str(w.message).startswith("Auth002")
    else:
        assert len(recwarn) == 0


def _get_access_token(url: str, user: str, pw: str) -> Dict[str, str]:
    # get an access token with WCS user and pw.
    weaviate_open_id_config = requests.get(url + "/v1/.well-known/openid-configuration")
    response_json = weaviate_open_id_config.json()
    client_id = response_json["clientId"]
    href = response_json["href"]

    # Get the token issuer's OIDC configuration
    response_auth = requests.get(href)

    # Construct the POST request to send to 'token_endpoint'
    auth_body = {
        "grant_type": "password",
        "client_id": client_id,
        "username": user,
        "password": pw,
        "scope": "openid offline_access",
    }
    response_post = requests.post(response_auth.json()["token_endpoint"], auth_body)
    return response_post.json()


@pytest.mark.parametrize(
    "name,user,env_variable_name,port",
    [
        (
            "WCS",
            "ms_2d0e007e7136de11d5f29fce7a53dae219a51458@existiert.net",
            "WCS_DUMMY_CI_PW",
            WCS_PORT,
        ),
        (
            "okta",
            "test@test.de",
            "OKTA_DUMMY_CI_PW",
            OKTA_PORT_USERS,
        ),
    ],
)
def test_authentication_with_bearer_token(name: str, user: str, env_variable_name: str, port: str):
    """Test authentication using existing bearer token."""
    url = f"http://localhost:{port}"
    assert is_auth_enabled(url)
    pw = os.environ.get(env_variable_name)
    if pw is None:
        pytest.skip(f"No login data for {name} found.")

    # use token to authenticate
    token = _get_access_token(url, user, pw)

    client = weaviate.Client(
        url,
        auth_client_secret=AuthBearerToken(
            access_token=token["access_token"],
            expires_in=int(token["expires_in"]),
            refresh_token=token["refresh_token"],
        ),
    )
    client.schema.delete_all()  # no exception


def test_client_with_authentication_with_anon_weaviate(recwarn):
    """Test that we warn users when their client has auth enabled, but weaviate has only anon access."""
    # testing for warnings can be flaky without this as there are open SSL conections
    warnings.filterwarnings(action="ignore", message="unclosed", category=ResourceWarning)

    url = f"http://localhost:{ANON_PORT}"
    assert not is_auth_enabled(url)

    client = weaviate.Client(
        url,
        auth_client_secret=AuthClientPassword(username="someUser", password="SomePw"),
    )

    # only one warning
    assert len(recwarn) == 1
    w = recwarn.pop()
    assert issubclass(w.category, UserWarning)
    assert str(w.message).startswith("Auth001")

    client.schema.delete_all()  # no exception, client works


def test_bearer_token_without_refresh(recwarn):
    """Test that the client warns users when only supplying an access token without refresh."""

    # testing for warnings can be flaky without this as there are open SSL conections
    warnings.filterwarnings(action="ignore", message="unclosed", category=ResourceWarning)

    url = f"http://localhost:{WCS_PORT}"
    assert is_auth_enabled(url)
    pw = os.environ.get("WCS_DUMMY_CI_PW")
    if pw is None:
        pytest.skip("No login data for WCS found.")

    token = _get_access_token(url, "ms_2d0e007e7136de11d5f29fce7a53dae219a51458@existiert.net", pw)
    client = weaviate.Client(
        url,
        auth_client_secret=AuthBearerToken(
            access_token=token["access_token"],
        ),
    )
    client.schema.delete_all()  # no exception, client works

    assert len(recwarn) == 1
    w = recwarn.pop()
    assert issubclass(w.category, UserWarning)
    assert str(w.message).startswith("Auth002")


def test_api_key():
    url = f"http://localhost:{WCS_PORT}"
    assert is_auth_enabled(url)

    client = weaviate.Client(url, auth_client_secret=AuthApiKey(api_key="my-secret-key"))
    client.schema.delete_all()  # no exception, client works


def test_api_key_wrong_key():
    url = f"http://localhost:{WCS_PORT}"
    assert is_auth_enabled(url)

    with pytest.raises(UnexpectedStatusCodeException) as e:
        weaviate.Client(url, auth_client_secret=AuthApiKey(api_key="wrong_key"))
        assert e.value.status_code == 401



"""
GraphQL `Aggregate` command.
"""
import json
from typing import List, Optional

from weaviate.connect import Connection
from weaviate.util import (
    _capitalize_first_letter,
    file_encoder_b64,
)
from .filter import (
    Where,
    GraphQL,
    Filter,
    NearAudio,
    NearDepth,
    NearImage,
    NearIMU,
    NearObject,
    NearText,
    NearThermal,
    NearVector,
    NearVideo,
    MediaType,
)


class AggregateBuilder(GraphQL):
    """
    AggregateBuilder class used to aggregate Weaviate objects.
    """

    def __init__(self, class_name: str, connection: Connection):
        """
        Initialize a AggregateBuilder class instance.

        Parameters
        ----------
        class_name : str
            Class name of the objects to be aggregated.
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.
        """

        super().__init__(connection)
        self._class_name: str = _capitalize_first_letter(class_name)
        self._object_limit: Optional[int] = None
        self._with_meta_count: bool = False
        self._fields: List[str] = []
        self._where: Optional[Where] = None
        self._group_by_properties: Optional[List[str]] = None
        self._uses_filter: bool = False
        self._near: Optional[Filter] = None
        self._tenant: Optional[str] = None
        self._limit: Optional[int] = None

    def with_tenant(self, tenant: str) -> "AggregateBuilder":
        """Sets a tenant for the query."""
        if not isinstance(tenant, str):
            raise TypeError("tenant must be of type str")

        self._tenant = tenant
        self._uses_filter = True
        return self

    def with_meta_count(self) -> "AggregateBuilder":
        """
        Set Meta Count to True.

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._with_meta_count = True
        return self

    def with_object_limit(self, limit: int) -> "AggregateBuilder":
        """
        Set objectLimit to limit vector search results used within the aggregation query
        only when with near<MEDIA> filter.

        Parameters
        ----------
        limit : int
            The object limit.

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._object_limit = limit
        return self

    def with_limit(self, limit: int) -> "AggregateBuilder":
        """
        Set limit to limit the number of returned results from the aggregation query.

        Parameters
        ----------
        limit : int
            The limit.

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._limit = limit
        return self

    def with_fields(self, field: str) -> "AggregateBuilder":
        """
        Include a field in the aggregate query.

        Parameters
        ----------
        field : str
            Field to include in the aggregate query.
            e.g. '<property_name> { count }'

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._fields.append(field)
        return self

    def with_where(self, content: dict) -> "AggregateBuilder":
        """
        Set 'where' filter.

        Parameters
        ----------
        content : dict
            The where filter to include in the aggregate query. See examples below.

        Examples
        --------
        The `content` prototype is like this:

        >>> content = {
        ...     'operator': '<operator>',
        ...     'operands': [
        ...         {
        ...             'path': [path],
        ...             'operator': '<operator>'
        ...             '<valueType>': <value>
        ...         },
        ...         {
        ...             'path': [<matchPath>],
        ...             'operator': '<operator>',
        ...             '<valueType>': <value>
        ...         }
        ...     ]
        ... }

        This is a complete `where` filter but it does not have to be like this all the time.

        Single operand:

        >>> content = {
        ...     'path': ["wordCount"],    # Path to the property that should be used
        ...     'operator': 'GreaterThan',  # operator
        ...     'valueInt': 1000       # value (which is always = to the type of the path property)
        ... }

        Or

        >>> content = {
        ...     'path': ["id"],
        ...     'operator': 'Equal',
        ...     'valueString': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf"
        ... }

        Multiple operands:

        >>> content = {
        ...     'operator': 'And',
        ...     'operands': [
        ...         {
        ...             'path': ["wordCount"],
        ...             'operator': 'GreaterThan',
        ...             'valueInt': 1000
        ...         },
        ...         {
        ...             'path': ["wordCount"],
        ...             'operator': 'LessThan',
        ...             'valueInt': 1500
        ...         }
        ...     ]
        ... }

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._where = Where(content)
        self._uses_filter = True
        return self

    def with_group_by_filter(self, properties: List[str]) -> "AggregateBuilder":
        """
        Add a group by filter to the query. Might requires the user to set
        an additional group by clause using `with_fields(..)`.

        Parameters
        ----------
        properties : list of str
            list of properties that are included in the group by filter.
            Generates a filter like: 'groupBy: ["property1", "property2"]'
            from a list ["property1", "property2"]

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._group_by_properties = properties
        self._uses_filter = True
        return self

    def with_near_text(self, content: dict) -> "AggregateBuilder":
        """
        Set `nearText` filter. This filter can be used with text modules (text2vec).
        E.g.: text2vec-contextionary, text2vec-transformers.
        NOTE: The 'autocorrect' field is enabled only with the `text-spellcheck` Weaviate module.

        Parameters
        ----------
        content : dict
            The content of the `nearText` filter to set. See examples below.

        Examples
        --------
        Content full prototype:

        >>> content = {
        ...     'concepts': <list of str or str>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'moveAwayFrom': { # Optional
        ...         'concepts': <list of str or str>,
        ...         'force': <float>
        ...     },
        ...     'moveTo': { # Optional
        ...         'concepts': <list of str or str>,
        ...         'force': <float>
        ...     },
        ...     'autocorrect': <bool>, # Optional
        ... }

        Full content:

        >>> content = {
        ...     'concepts': ["fashion"],
        ...     'certainty': 0.7, # or 'distance' instead
        ...     'moveAwayFrom': {
        ...         'concepts': ["finance"],
        ...         'force': 0.45
        ...     },
        ...     'moveTo': {
        ...         'concepts': ["haute couture"],
        ...         'force': 0.85
        ...     },
        ...     'autocorrect': True
        ... }

        Partial content:

        >>> content = {
        ...     'concepts': ["fashion"],
        ...     'certainty': 0.7, # or 'distance' instead
        ...     'moveTo': {
        ...         'concepts': ["haute couture"],
        ...         'force': 0.85
        ...     }
        ... }

        Minimal content:

        >>> content = {
        ...     'concepts': "fashion"
        ... }

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        if self._near is not None:
            raise AttributeError("Cannot use multiple 'near' filters.")
        self._near = NearText(content)
        self._uses_filter = True
        return self

    def with_near_vector(self, content: dict) -> "AggregateBuilder":
        """
        Set `nearVector` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearVector` filter to set. See examples below.

        Examples
        --------
        Content full prototype:

        >>> content = {
        ...     'vector' : <list of float>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        NOTE: Supported types for 'vector' are `list`, 'numpy.ndarray`, `torch.Tensor`
                and `tf.Tensor`.

        Full content:

        >>> content = {
        ...     'vector' : [.1, .2, .3, .5],
        ...     'certainty': 0.75, # or 'distance' instead
        ... }

        Minimal content:

        >>> content = {
        ...     'vector' : [.1, .2, .3, .5]
        ... }

        Or

        >>> content = {
        ...     'vector' : torch.tensor([.1, .2, .3, .5])
        ... }

        Or

        >>> content = {
        ...     'vector' : torch.tensor([[.1, .2, .3, .5]]) # it is going to be squeezed.
        ... }

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        if self._near is not None:
            raise AttributeError("Cannot use multiple 'near' filters.")
        self._near = NearVector(content)
        self._uses_filter = True
        return self

    def with_near_object(self, content: dict) -> "AggregateBuilder":
        """
        Set `nearObject` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearObject` filter to set. See examples below.

        Examples
        --------
        Content prototype:

        >>> content = {
        ...     'id': <str>, # OR 'beacon'
        ...     'beacon': <str>, # OR 'id'
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        >>> {
        ...     'id': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> # alternatively
        >>> {
        ...     'beacon': "weaviate://localhost/Book/e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf"
        ...     'certainty': 0.7 # or 'distance' instead
        ... }

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        is_server_version_14 = self._connection.server_version >= "1.14"

        if self._near is not None:
            raise AttributeError("Cannot use multiple 'near' filters.")
        self._near = NearObject(content, is_server_version_14)
        self._uses_filter = True
        return self

    def with_near_image(self, content: dict, encode: bool = True) -> "AggregateBuilder":
        """
        Set `nearImage` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearImage` filter to set. See examples below.
        encode : bool, optional
            Whether to encode the `content["image"]` to base64 and convert to string. If True, the
            `content["image"]` can be an image path or a file opened in binary read mode. If False,
            the `content["image"]` MUST be a base64 encoded string (NOT bytes, i.e. NOT binary
            string that looks like this: b'BASE64ENCODED' but simple 'BASE64ENCODED').
            By default True.

        Examples
        --------
        Content prototype:

        >>> content = {
        ...     'image': <str or binary read file>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        >>> {
        ...     'image': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     'certainty': 0.7 # or 'distance'
        ... }

        With `encoded` True:

        >>> content = {
        ...     'image': "my_image_path.png",
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Image')\\
                .with_fields('description')\\
        ...     .with_near_image(content, encode=True) # <- encode MUST be set to True

        OR

        >>> my_image_file = open("my_image_path.png", "br")
        >>> content = {
        ...     'image': my_image_file,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Image')\\
                .with_fields('description')\\
        ...     .with_near_image(content, encode=True) # <- encode MUST be set to True
        >>> my_image_file.close()

        With `encoded` False:

        >>> from weaviate.util import image_encoder_b64, image_decoder_b64
        >>> encoded_image = image_encoder_b64("my_image_path.png")
        >>> content = {
        ...     'image': encoded_image,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Image')\\
                .with_fields('description')\\
        ...     .with_near_image(content, encode=False) # <- encode MUST be set to False

        OR

        >>> from weaviate.util import image_encoder_b64, image_decoder_b64
        >>> with open("my_image_path.png", "br") as my_image_file:
        ...     encoded_image = image_encoder_b64(my_image_file)
        >>> content = {
        ...     'image': encoded_image,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Image')\\
                .with_fields('description')\\
        ...     .with_near_image(content, encode=False) # <- encode MUST be set to False

        Encode Image yourself:

        >>> import base64
        >>> with open("my_image_path.png", "br") as my_image_file:
        ...     encoded_image = base64.b64encode(my_image_file.read()).decode("utf-8")
        >>> content = {
        ...     'image': encoded_image,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Image')\\
                .with_fields('description')\\
        ...     .with_near_image(content, encode=False) # <- encode MUST be set to False

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            The updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """
        self._media_type = MediaType.IMAGE
        if self._near is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        if encode:
            content["image"] = file_encoder_b64(content["image"])
        self._near = NearImage(content)
        self._uses_filter = True
        return self

    def with_near_audio(self, content: dict, encode: bool = True) -> "AggregateBuilder":
        """
        Set `nearAudio` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearAudio` filter to set. See examples below.
        encode : bool, optional
            Whether to encode the `content["audio"]` to base64 and convert to string. If True, the
            `content["audio"]` can be an audio path or a file opened in binary read mode. If False,
            the `content["audio"]` MUST be a base64 encoded string (NOT bytes, i.e. NOT binary
            string that looks like this: b'BASE64ENCODED' but simple 'BASE64ENCODED').
            By default True.

        Examples
        --------
        Content prototype:

        >>> content = {
        ...     'audio': <str or binary read file>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        >>> {
        ...     'audio': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     'certainty': 0.7 # or 'distance'
        ... }

        With `encoded` True:

        >>> content = {
        ...     'audio': "my_audio_path.wav",
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Audio')\\
                .with_fields('description')\\
        ...     .with_near_audio(content, encode=True) # <- encode MUST be set to True

        OR

        >>> my_audio_file = open("my_audio_path.wav", "br")
        >>> content = {
        ...     'audio': my_audio_file,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Audio')\\
                .with_fields('description')\\
        ...     .with_near_audio(content, encode=True) # <- encode MUST be set to True
        >>> my_audio_file.close()

        With `encoded` False:

        >>> from weaviate.util import file_encoder_b64
        >>> encoded_audio = file_encoder_b64("my_audio_path.wav")
        >>> content = {
        ...     'audio': encoded_audio,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Audio')\\
                .with_fields('description')\\
        ...     .with_near_audio(content, encode=False) # <- encode MUST be set to False

        OR

        >>> from weaviate.util import file_encoder_b64
        >>> with open("my_audio_path.wav", "br") as my_audio_file:
        ...     encoded_audio = file_encoder_b64(my_audio_file)
        >>> content = {
        ...     'audio': encoded_audio,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Audio')\\
                .with_fields('description')\\
        ...     .with_near_audio(content, encode=False) # <- encode MUST be set to False

        Encode Audio yourself:

        >>> import base64
        >>> with open("my_audio_path.wav", "br") as my_audio_file:
        ...     encoded_audio = base64.b64encode(my_audio_file.read()).decode("utf-8")
        >>> content = {
        ...     'audio': encoded_audio,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Audio')\\
                .with_fields('description')\\
        ...     .with_near_audio(content, encode=False) # <- encode MUST be set to False

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            The updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        self._media_type = MediaType.AUDIO
        if self._near is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        if encode:
            content[self._media_type.value] = file_encoder_b64(content[self._media_type.value])
        self._near = NearAudio(content)
        self._uses_filter = True
        return self

    def with_near_video(self, content: dict, encode: bool = True) -> "AggregateBuilder":
        """
        Set `nearVideo` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearVideo` filter to set. See examples below.
        encode : bool, optional
            Whether to encode the `content["video"]` to base64 and convert to string. If True, the
            `content["video"]` can be an video path or a file opened in binary read mode. If False,
            the `content["video"]` MUST be a base64 encoded string (NOT bytes, i.e. NOT binary
            string that looks like this: b'BASE64ENCODED' but simple 'BASE64ENCODED').
            By default True.

        Examples
        --------
        Content prototype:

        >>> content = {
        ...     'video': <str or binary read file>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        >>> {
        ...     'video': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     'certainty': 0.7 # or 'distance'
        ... }

        With `encoded` True:

        >>> content = {
        ...     'video': "my_video_path.avi",
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Video')\\
                .with_fields('description')\\
        ...     .with_near_video(content, encode=True) # <- encode MUST be set to True

        OR

        >>> my_video_file = open("my_video_path.avi", "br")
        >>> content = {
        ...     'video': my_video_file,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Video')\\
                .with_fields('description')\\
        ...     .with_near_video(content, encode=True) # <- encode MUST be set to True
        >>> my_video_file.close()

        With `encoded` False:

        >>> from weaviate.util import file_encoder_b64
        >>> encoded_video = file_encoder_b64("my_video_path.avi")
        >>> content = {
        ...     'video': encoded_video,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Video')\\
                .with_fields('description')\\
        ...     .with_near_video(content, encode=False) # <- encode MUST be set to False

        OR

        >>> from weaviate.util import file_encoder_b64, video_decoder_b64
        >>> with open("my_video_path.avi", "br") as my_video_file:
        ...     encoded_video = file_encoder_b64(my_video_file)
        >>> content = {
        ...     'video': encoded_video,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Video')\\
                .with_fields('description')\\
        ...     .with_near_video(content, encode=False) # <- encode MUST be set to False

        Encode Video yourself:

        >>> import base64
        >>> with open("my_video_path.avi", "br") as my_video_file:
        ...     encoded_video = base64.b64encode(my_video_file.read()).decode("utf-8")
        >>> content = {
        ...     'video': encoded_video,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Video')\\
                .with_fields('description')\\
        ...     .with_near_video(content, encode=False) # <- encode MUST be set to False

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            The updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        self._media_type = MediaType.VIDEO
        if self._near is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        if encode:
            content[self._media_type.value] = file_encoder_b64(content[self._media_type.value])
        self._near = NearVideo(content)
        self._uses_filter = True
        return self

    def with_near_depth(self, content: dict, encode: bool = True) -> "AggregateBuilder":
        """
        Set `nearDepth` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearDepth` filter to set. See examples below.
        encode : bool, optional
            Whether to encode the `content["depth"]` to base64 and convert to string. If True, the
            `content["depth"]` can be an depth path or a file opened in binary read mode. If False,
            the `content["depth"]` MUST be a base64 encoded string (NOT bytes, i.e. NOT binary
            string that looks like this: b'BASE64ENCODED' but simple 'BASE64ENCODED').
            By default True.

        Examples
        --------
        Content prototype:

        >>> content = {
        ...     'depth': <str or binary read file>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        >>> {
        ...     'depth': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     'certainty': 0.7 # or 'distance'
        ... }

        With `encoded` True:

        >>> content = {
        ...     'depth': "my_depth_path.png",
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Depth')\\
                .with_fields('description')\\
        ...     .with_near_depth(content, encode=True) # <- encode MUST be set to True

        OR

        >>> my_depth_file = open("my_depth_path.png", "br")
        >>> content = {
        ...     'depth': my_depth_file,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Depth')\\
                .with_fields('description')\\
        ...     .with_near_depth(content, encode=True) # <- encode MUST be set to True
        >>> my_depth_file.close()

        With `encoded` False:

        >>> from weaviate.util import file_encoder_b64
        >>> encoded_depth = file_encoder_b64("my_depth_path.png")
        >>> content = {
        ...     'depth': encoded_depth,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Depth')\\
                .with_fields('description')\\
        ...     .with_near_depth(content, encode=False) # <- encode MUST be set to False

        OR

        >>> from weaviate.util import file_encoder_b64
        >>> with open("my_depth_path.png", "br") as my_depth_file:
        ...     encoded_depth = file_encoder_b64(my_depth_file)
        >>> content = {
        ...     'depth': encoded_depth,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Depth')\\
                .with_fields('description')\\
        ...     .with_near_depth(content, encode=False) # <- encode MUST be set to False

        Encode Depth yourself:

        >>> import base64
        >>> with open("my_depth_path.png", "br") as my_depth_file:
        ...     encoded_depth = base64.b64encode(my_depth_file.read()).decode("utf-8")
        >>> content = {
        ...     'depth': encoded_depth,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Depth')\\
                .with_fields('description')\\
        ...     .with_near_depth(content, encode=False) # <- encode MUST be set to False

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            The updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        self._media_type = MediaType.DEPTH
        if self._near is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        if encode:
            content[self._media_type.value] = file_encoder_b64(content[self._media_type.value])
        self._near = NearDepth(content)
        self._uses_filter = True
        return self

    def with_near_thermal(self, content: dict, encode: bool = True) -> "AggregateBuilder":
        """
        Set `nearThermal` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearThermal` filter to set. See examples below.
        encode : bool, optional
            Whether to encode the `content["thermal"]` to base64 and convert to string. If True, the
            `content["thermal"]` can be an thermal path or a file opened in binary read mode. If False,
            the `content["thermal"]` MUST be a base64 encoded string (NOT bytes, i.e. NOT binary
            string that looks like this: b'BASE64ENCODED' but simple 'BASE64ENCODED').
            By default True.

        Examples
        --------
        Content prototype:

        >>> content = {
        ...     'thermal': <str or binary read file>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        >>> {
        ...     'thermal': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     'certainty': 0.7 # or 'distance'
        ... }

        With `encoded` True:

        >>> content = {
        ...     'thermal': "my_thermal_path.png",
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.get('Thermal', 'description')\\
        ...     .with_near_thermal(content, encode=True) # <- encode MUST be set to True

        OR

        >>> my_thermal_file = open("my_thermal_path.png", "br")
        >>> content = {
        ...     'thermal': my_thermal_file,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Thermal')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=True) # <- encode MUST be set to True
        >>> my_thermal_file.close()

        With `encoded` False:

        >>> from weaviate.util import file_encoder_b64
        >>> encoded_thermal = file_encoder_b64("my_thermal_path.png")
        >>> content = {
        ...     'thermal': encoded_thermal,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Thermal')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=False) # <- encode MUST be set to False

        OR

        >>> from weaviate.util import file_encoder_b64
        >>> with open("my_thermal_path.png", "br") as my_thermal_file:
        ...     encoded_thermal = file_encoder_b64(my_thermal_file)
        >>> content = {
        ...     'thermal': encoded_thermal,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Thermal')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=False) # <- encode MUST be set to False

        Encode Thermal yourself:

        >>> import base64
        >>> with open("my_thermal_path.png", "br") as my_thermal_file:
        ...     encoded_thermal = base64.b64encode(my_thermal_file.read()).decode("utf-8")
        >>> content = {
        ...     'thermal': encoded_thermal,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('Thermal')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=False) # <- encode MUST be set to False

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            The updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        self._media_type = MediaType.THERMAL
        if self._near is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        if encode:
            content[self._media_type.value] = file_encoder_b64(content[self._media_type.value])
        self._near = NearThermal(content)
        self._uses_filter = True
        return self

    def with_near_imu(self, content: dict, encode: bool = True) -> "AggregateBuilder":
        """
        Set `nearIMU` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearIMU` filter to set. See examples below.
        encode : bool, optional
            Whether to encode the `content["thermal"]` to base64 and convert to string. If True, the
            `content["thermal"]` can be an thermal path or a file opened in binary read mode. If False,
            the `content["thermal"]` MUST be a base64 encoded string (NOT bytes, i.e. NOT binary
            string that looks like this: b'BASE64ENCODED' but simple 'BASE64ENCODED').
            By default True.

        Examples
        --------
        Content prototype:

        >>> content = {
        ...     'thermal': <str or binary read file>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        >>> {
        ...     'thermal': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     'certainty': 0.7 # or 'distance'
        ... }

        With `encoded` True:

        >>> content = {
        ...     'thermal': "my_thermal_path.png",
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('IMU')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=True) # <- encode MUST be set to True

        OR

        >>> my_thermal_file = open("my_thermal_path.png", "br")
        >>> content = {
        ...     'thermal': my_thermal_file,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('IMU')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=True) # <- encode MUST be set to True
        >>> my_thermal_file.close()

        With `encoded` False:

        >>> from weaviate.util import file_encoder_b64
        >>> encoded_thermal = file_encoder_b64("my_thermal_path.png")
        >>> content = {
        ...     'thermal': encoded_thermal,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('IMU')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=False) # <- encode MUST be set to False

        OR

        >>> from weaviate.util import file_encoder_b64
        >>> with open("my_thermal_path.png", "br") as my_thermal_file:
        ...     encoded_thermal = file_encoder_b64(my_thermal_file)
        >>> content = {
        ...     'thermal': encoded_thermal,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('IMU')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=False) # <- encode MUST be set to False

        Encode IMU yourself:

        >>> import base64
        >>> with open("my_thermal_path.png", "br") as my_thermal_file:
        ...     encoded_thermal = base64.b64encode(my_thermal_file.read()).decode("utf-8")
        >>> content = {
        ...     'thermal': encoded_thermal,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.aggregate('IMU')\\
        ...     .with_fields('description')\\
        ...     .with_near_thermal(content, encode=False) # <- encode MUST be set to False

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            The updated AggregateBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        self._media_type = MediaType.IMU
        if self._near is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        if encode:
            content[self._media_type.value] = file_encoder_b64(content[self._media_type.value])
        self._near = NearIMU(content)
        self._uses_filter = True
        return self

    def build(self) -> str:
        """
        Build the query and return the string.

        Returns
        -------
        str
            The GraphQL query as a string.
        """

        # Path
        query = f"{{Aggregate{{{self._class_name}"

        # Filter
        if self._uses_filter:
            query += "("
            if self._where is not None:
                query += str(self._where)
            if self._group_by_properties is not None:
                query += f"groupBy: {json.dumps(self._group_by_properties)}"
            if self._near is not None:
                query += str(self._near)
            if self._object_limit:
                query += f"objectLimit: {self._object_limit}"
            if self._tenant is not None:
                query += f'tenant: "{self._tenant}"'
            if self._limit is not None:
                query += f"limit: {self._limit}"

            query += ")"

        # Body
        query += "{"
        if self._with_meta_count:
            query += "meta{count}"
        for field in self._fields:
            query += field

        # close
        query += "}}}"
        return query



"""
Cluster class definition.
"""
from typing import List, Optional, cast

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.cluster.types import Node
from weaviate.connect import Connection
from weaviate.exceptions import (
    EmptyResponseException,
)
from ..util import _capitalize_first_letter, _decode_json_response_dict


class Cluster:
    """
    Cluster class used for cluster information
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Cluster class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.
        """

        self._connection = connection

    def get_nodes_status(self, class_name: Optional[str] = None) -> List[Node]:
        """
        Get the nodes status.

        Parameters
        ----------
        class_name : Optional[str]
            Get the status for the given class. If not given all classes will be included.

        Returns
        -------
        list
            List of nodes and their respective status.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        weaviate.EmptyResponseException
            If the response is empty.
        """
        path = "/nodes"
        if class_name is not None:
            path += "/" + _capitalize_first_letter(class_name)

        try:
            response = self._connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Get nodes status failed due to connection error"
            ) from conn_err

        response_typed = _decode_json_response_dict(response, "Nodes status")
        assert response_typed is not None
        nodes = response_typed.get("nodes")
        if nodes is None or nodes == []:
            raise EmptyResponseException("Nodes status response returned empty")
        return cast(List[Node], nodes)



from typing import Generic, List, Optional, Type, overload

from weaviate.collections.classes.filters import (
    _Filters,
)
from weaviate.collections.classes.grpc import MetadataQuery, PROPERTIES, HybridFusion
from weaviate.collections.classes.internal import (
    _GenerativeReturn,
    _QueryReturn,
    _Generative,
    GenerativeReturn,
    QueryReturn,
    ReturnProperties,
)
from weaviate.collections.classes.types import Properties, TProperties
from weaviate.collections.queries.base import _Grpc


class _HybridQuery(Generic[Properties], _Grpc[Properties]):
    @overload
    def hybrid(
        self,
        query: str,
        alpha: Optional[float] = None,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> _QueryReturn[Properties]:
        ...

    @overload
    def hybrid(
        self,
        query: str,
        alpha: Optional[float] = None,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        *,
        return_properties: Type[TProperties],
    ) -> _QueryReturn[TProperties]:
        ...

    def hybrid(
        self,
        query: str,
        alpha: Optional[float] = None,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
    ) -> QueryReturn[Properties, TProperties]:
        """Search for objects in this collection using the hybrid algorithm blending keyword-based BM25 and vector-based similarity.

        See the [docs](https://weaviate.io/developers/weaviate/search/hybrid) for a more detailed explanation.

        Arguments:
            `query`
                The keyword-based query to search for, REQUIRED.
            `alpha`
                The weight of the BM25 score. If not specified, the default weight specified by the server is used.
            `vector`
                The specific vector to search for. If not specified, the query is vectorised and used in the similarity search.
            `query_properties`
                The properties to search in. If not specified, all properties are searched.
            `fusion_type`
                The type of fusion to apply. If not specified, the default fusion type specified by the server is used.
            `limit`
                The maximum number of results to return. If not specified, the default limit specified by the server is returned.
            `auto_limit`
                The maximum number of [autocut](https://weaviate.io/developers/weaviate/api/graphql/additional-operators#autocut) results to return. If not specified, no limit is applied.
            `filters`
                The filters to apply to the search.
            `return_metadata`
                The metadata to return for each object.
            `return_properties`
                The properties to return for each object.

        NOTE:
            If neither `return_metadata` nor `return_properties` are provided then all properties and metadata are returned except for `metadata.vector`.

        Returns:
            A `_QueryReturn` object that includes the searched objects.

        Raises:
            `weaviate.exceptions.WeaviateQueryException`:
                If the network connection to Weaviate fails.
        """
        ret_properties, ret_metadata = self._parse_return_properties(return_properties)
        res = self._query().hybrid(
            query=query,
            alpha=alpha,
            vector=vector,
            properties=query_properties,
            fusion_type=fusion_type,
            limit=limit,
            autocut=auto_limit,
            filters=filters,
            return_metadata=return_metadata or ret_metadata,
            return_properties=ret_properties,
        )
        return self._result_to_query_return(res, return_properties)


class _HybridGenerate(Generic[Properties], _Grpc[Properties]):
    @overload
    def hybrid(
        self,
        query: str,
        single_prompt: Optional[str] = None,
        grouped_task: Optional[str] = None,
        grouped_properties: Optional[List[str]] = None,
        alpha: Optional[float] = None,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> _GenerativeReturn[Properties]:
        ...

    @overload
    def hybrid(
        self,
        query: str,
        single_prompt: Optional[str] = None,
        grouped_task: Optional[str] = None,
        grouped_properties: Optional[List[str]] = None,
        alpha: Optional[float] = None,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        *,
        return_properties: Type[TProperties],
    ) -> _GenerativeReturn[TProperties]:
        ...

    def hybrid(
        self,
        query: str,
        single_prompt: Optional[str] = None,
        grouped_task: Optional[str] = None,
        grouped_properties: Optional[List[str]] = None,
        alpha: Optional[float] = None,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
    ) -> GenerativeReturn[Properties, TProperties]:
        """Perform retrieval-augmented generation (RaG) on the results of an object search in this collection using the hybrid algorithm blending keyword-based BM25 and vector-based similarity.

        See the [docs](https://weaviate.io/developers/weaviate/search/hybrid) for a more detailed explanation.

        Arguments:
            `query`
                The keyword-based query to search for, REQUIRED.
            `single_prompt`
                The prompt to use for RaG on each object individually.
            `grouped_task`
                The prompt to use for RaG on the entire result set.
            `grouped_properties`
                The properties to use in the RaG on the entire result set.
            `alpha`
                The weight of the BM25 score. If not specified, the default weight specified by the server is used.
            `vector`
                The specific vector to search for. If not specified, the query is vectorised and used in the similarity search.
            `query_properties`
                The properties to search in. If not specified, all properties are searched.
            `fusion_type`
                The type of fusion to apply. If not specified, the default fusion type specified by the server is used.
            `limit`
                The maximum number of results to return. If not specified, the default limit specified by the server is returned.
            `auto_limit`
                The maximum number of [autocut](https://weaviate.io/developers/weaviate/api/graphql/additional-operators#autocut) results to return. If not specified, no limit is applied.
            `filters`
                The filters to apply to the search.
            `return_metadata`
                The metadata to return for each object.
            `return_properties`
                The properties to return for each object.

        NOTE:
            If neither `return_metadata` nor `return_properties` are provided then all properties and metadata are returned except for `metadata.vector`.

        Returns:
            A `_GenerativeReturn` object that includes the searched objects with per-object generated results and group generated results.

        Raises:
            `weaviate.exceptions.WeaviateQueryException`:
                If the network connection to Weaviate fails.
        """
        ret_properties, ret_metadata = self._parse_return_properties(return_properties)
        res = self._query().hybrid(
            query=query,
            alpha=alpha,
            vector=vector,
            properties=query_properties,
            fusion_type=fusion_type,
            limit=limit,
            autocut=auto_limit,
            filters=filters,
            return_metadata=return_metadata or ret_metadata,
            return_properties=ret_properties,
            generative=_Generative(
                single=single_prompt,
                grouped=grouped_task,
                grouped_properties=grouped_properties,
            ),
        )
        return self._result_to_generative_return(res, return_properties)



import unittest
from typing import List, Optional, Callable, Tuple
from unittest.mock import patch, Mock

import pytest

from test.util import check_error_message
from weaviate.data.replication import ConsistencyLevel
from weaviate.gql.get import (
    GetBuilder,
    BM25,
    Hybrid,
    LinkTo,
    GroupBy,
    AdditionalProperties,
    HybridFusion,
)

mock_connection_v117 = Mock()
mock_connection_v117.server_version = "1.17.4"


@pytest.mark.parametrize(
    "props,expected",
    [
        (AdditionalProperties(uuid=True), " _additional{id} "),
        (
            AdditionalProperties(uuid=True, vector=True, explainScore=True),
            " _additional{id vector explainScore} ",
        ),
        (
            AdditionalProperties(uuid=False, vector=True, explainScore=True, score=True),
            " _additional{vector score explainScore} ",
        ),
    ],
)
def test_additional_props(props: AdditionalProperties, expected: str):
    assert str(props) == expected


@pytest.mark.parametrize(
    "additional_props,expected",
    [
        (AdditionalProperties(uuid=True), " _additional{id} "),
    ],
)
def test_getbuilder_with_additional_props(additional_props: AdditionalProperties, expected: str):
    query = (
        GetBuilder("TestClass", "name", mock_connection_v117)
        .with_additional(additional_props)
        .build()
    )
    expected_query = "{Get{TestClass{name" + expected + "}}}"
    assert (
        "name_" not in expected_query
    )  # Check that the prop name is not being concatnated to _additional
    assert query == expected_query


@pytest.mark.parametrize(
    "query,properties,expected",
    [
        (
            "query",
            ["title", "document", "date"],
            'bm25:{query: "query", properties: ["title","document","date"]}',
        ),
        ("other query", [], 'bm25:{query: "other query"}'),
        ("other query", None, 'bm25:{query: "other query"}'),
        ('what is an "airport"', None, 'bm25:{query: "what is an \\"airport\\""}'),
        ("what is an 'airport'", None, """bm25:{query: "what is an 'airport'"}"""),
    ],
)
def test_bm25(query: str, properties: List[str], expected: str):
    bm25 = BM25(query, properties)
    assert str(bm25) == expected


@pytest.mark.parametrize(
    "property_name,in_class,properties,expected",
    [
        (
            "property",
            "class",
            ["title"],
            "property{... on class{title}}",
        ),
        (
            "property",
            "class",
            ["title", "document", "date"],
            "property{... on class{title document date}}",
        ),
    ],
)
def test_get_references(property_name: str, in_class: str, properties: List[str], expected: str):
    ref = LinkTo(property_name, in_class, properties)
    assert str(ref) == expected


@pytest.mark.parametrize(
    "query,vector,alpha,properties,fusion_type,expected",
    [
        (
            "query",
            [1, 2, 3],
            0.5,
            None,
            None,
            'hybrid:{query: "query", vector: [1, 2, 3], alpha: 0.5}',
        ),
        ("query", None, None, None, None, 'hybrid:{query: "query"}'),
        ('query "query2"', None, None, None, None, 'hybrid:{query: "query \\"query2\\""}'),
        ("query 'query2'", None, None, None, None, """hybrid:{query: "query 'query2'"}"""),
        ("query", None, None, ["prop1"], None, 'hybrid:{query: "query", properties: ["prop1"]}'),
        (
            "query",
            None,
            None,
            ["prop1", "prop2"],
            None,
            'hybrid:{query: "query", properties: ["prop1","prop2"]}',
        ),
        (
            "query",
            None,
            None,
            None,
            HybridFusion.RANKED,
            'hybrid:{query: "query", fusionType: rankedFusion}',
        ),
        (
            "query",
            None,
            None,
            None,
            HybridFusion.RELATIVE_SCORE,
            'hybrid:{query: "query", fusionType: relativeScoreFusion}',
        ),
        (
            "query",
            None,
            None,
            None,
            "relativeScoreFusion",
            'hybrid:{query: "query", fusionType: relativeScoreFusion}',
        ),
    ],
)
def test_hybrid(
    query: str,
    vector: Optional[List[float]],
    alpha: Optional[float],
    properties: Optional[List[str]],
    fusion_type: HybridFusion,
    expected: str,
):
    hybrid = Hybrid(query, alpha, vector, properties, fusion_type)
    assert str(hybrid) == expected


@pytest.mark.parametrize(
    "single_prompt,grouped_task,grouped_properties,expected",
    [
        (
            "What is the meaning of life?",
            None,
            None,
            """generate(singleResult:{prompt:"What is the meaning of life?"}){error singleResult} """,
        ),
        (
            'What is the meaning of "life"?',
            None,
            None,
            """generate(singleResult:{prompt:"What is the meaning of \\"life\\"?"}){error singleResult} """,
        ),
        (
            "What is the meaning of 'life'?",
            None,
            None,
            """generate(singleResult:{prompt:"What is the meaning of 'life'?"}){error singleResult} """,
        ),
        (
            None,
            "Explain why these magazines or newspapers are about finance",
            None,
            """generate(groupedResult:{task:"Explain why these magazines or newspapers are about finance"}){error groupedResult} """,
        ),
        (
            None,
            'Explain why these magazines or newspapers are about "finance"',
            None,
            """generate(groupedResult:{task:"Explain why these magazines or newspapers are about \\"finance\\""}){error groupedResult} """,
        ),
        (
            None,
            "Explain why these magazines or newspapers are about 'finance'",
            None,
            """generate(groupedResult:{task:"Explain why these magazines or newspapers are about 'finance'"}){error groupedResult} """,
        ),
        (
            "What is the meaning of life?",
            "Explain why these magazines or newspapers are about finance",
            None,
            """generate(singleResult:{prompt:"What is the meaning of life?"}groupedResult:{task:"Explain why these magazines or newspapers are about finance"}){error singleResult groupedResult} """,
        ),
        (
            None,
            "Explain why these magazines or newspapers are about finance",
            ["description"],
            """generate(groupedResult:{task:"Explain why these magazines or newspapers are about finance",properties:["description"]}){error groupedResult} """,
        ),
        (
            "What is the meaning of life?",
            "Explain why these magazines or newspapers are about finance",
            ["title", "description"],
            """generate(singleResult:{prompt:"What is the meaning of life?"}groupedResult:{task:"Explain why these magazines or newspapers are about finance",properties:["title","description"]}){error singleResult groupedResult} """,
        ),
    ],
)
def test_generative(
    single_prompt: str, grouped_task: str, grouped_properties: List[str], expected: str
):
    query = (
        GetBuilder("Person", "name", mock_connection_v117)
        .with_generate(single_prompt, grouped_task, grouped_properties)
        .build()
    )
    expected_query = "{Get{Person{name _additional {" + expected + "}}}}"
    assert query == expected_query


@pytest.mark.parametrize("single_prompt,grouped_task", [(123, None), (None, None), (None, 123)])
def test_generative_type(single_prompt: str, grouped_task: str):
    with pytest.raises(TypeError):
        GetBuilder("Person", "name", mock_connection_v117).with_generate(
            single_prompt, grouped_task
        ).build()


@pytest.mark.parametrize(
    "properties,groups,max_groups,expected",
    [
        (
            ["prop1", "prop2"],
            2,
            3,
            'groupBy:{path:["prop1","prop2"], groups:2, objectsPerGroup:3}',
        ),
        (
            ["prop1"],
            4,
            5,
            'groupBy:{path:["prop1"], groups:4, objectsPerGroup:5}',
        ),
    ],
)
def test_groupy(properties: List[str], groups: int, max_groups: int, expected: str):
    assert str(GroupBy(properties, groups, max_groups)) == expected


class TestGetBuilder(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        class_name_error_msg = f"class name must be of type str but was {int}"
        properties_error_msg = (
            "properties must be of type str, " f"list of str or None but was {int}"
        )
        property_error_msg = "All the `properties` must be of type `str` or Reference!"

        # invalid calls
        with self.assertRaises(TypeError) as error:
            GetBuilder(1, ["a"], None)
        check_error_message(self, error, class_name_error_msg)

        with self.assertRaises(TypeError) as error:
            GetBuilder("A", 2, None)
        check_error_message(self, error, properties_error_msg)

        with self.assertRaises(TypeError) as error:
            GetBuilder("A", [True], None)
        check_error_message(self, error, property_error_msg)

        # valid calls
        GetBuilder("name", "prop", None)
        GetBuilder("name", ["prop1", "prop2"], None)

    def test_build_with_limit(self):
        """
        Test the `with_limit` method.
        """

        # valid calls
        query = GetBuilder("Person", "name", None).with_limit(20).build()
        self.assertEqual("{Get{Person(limit: 20 ){name}}}", query)

        # invalid calls
        limit_error_msg = "limit cannot be non-positive (limit >=1)."
        with self.assertRaises(ValueError) as error:
            GetBuilder("A", ["str"], None).with_limit(0)
        check_error_message(self, error, limit_error_msg)

        with self.assertRaises(ValueError) as error:
            GetBuilder("A", ["str"], None).with_limit(-1)
        check_error_message(self, error, limit_error_msg)

    def test_build_with_offset(self):
        """
        Test the `with_limit` method.
        """

        # valid calls
        query = GetBuilder("Person", "name", None).with_offset(20).build()
        self.assertEqual("{Get{Person(offset: 20 ){name}}}", query)

        query = GetBuilder("Person", "name", None).with_offset(0).build()
        self.assertEqual("{Get{Person(offset: 0 ){name}}}", query)

        # invalid calls
        limit_error_msg = "offset cannot be non-positive (offset >=0)."
        with self.assertRaises(ValueError) as error:
            GetBuilder("A", ["str"], None).with_offset(-1)
        check_error_message(self, error, limit_error_msg)

    def test_build_with_consistency_level(self):
        """
        Test the `with_consistency_level` method
        """

        query = (
            GetBuilder("Person", "name", None).with_consistency_level(ConsistencyLevel.ONE).build()
        )
        self.assertEqual("{Get{Person(consistencyLevel: ONE ){name}}}", query)

        query = (
            GetBuilder("Person", "name", None)
            .with_consistency_level(ConsistencyLevel.QUORUM)
            .build()
        )
        self.assertEqual("{Get{Person(consistencyLevel: QUORUM ){name}}}", query)

        query = (
            GetBuilder("Person", "name", None).with_consistency_level(ConsistencyLevel.ALL).build()
        )
        self.assertEqual("{Get{Person(consistencyLevel: ALL ){name}}}", query)

    def test_build_with_where(self):
        """
        Test the ` with_where` method.
        """

        filter_name = {"path": ["name"], "operator": "Equal", "valueString": "A"}
        query = GetBuilder("Person", "name", None).with_where(filter_name).build()
        self.assertEqual(
            '{Get{Person(where: {path: ["name"] operator: Equal valueString: "A"} ){name}}}', query
        )

    def test_build_with_near_text(self):
        """
        Test the `with_near_text` method.
        """

        near_text = {
            "concepts": "computer",
            "moveTo": {"concepts": ["science"], "force": 0.5},
            "autocorrect": True,
        }

        # valid calls
        query = GetBuilder("Person", "name", None).with_near_text(near_text).build()
        self.assertEqual(
            '{Get{Person(nearText: {concepts: ["computer"] moveTo: {force: 0.5 concepts: ["science"]} autocorrect: true} ){name}}}',
            query,
        )

        # invalid calls
        near_error_msg = (
            "Cannot use multiple 'near' filters, or a 'near' filter along with a 'ask' filter!"
        )

        near_vector = {"vector": [1, 2, 3, 4, 5, 6, 7, 8, 9], "certainty": 0.55}
        with self.assertRaises(AttributeError) as error:
            GetBuilder("Person", "name", None).with_near_vector(near_vector).with_near_text(
                near_text
            )
        check_error_message(self, error, near_error_msg)

    def test_build_near_vector(self):
        """
        Test the `with_near_vector` method.
        """

        near_vector = {"vector": [1, 2, 3, 4, 5, 6, 7, 8, 9], "certainty": 0.55}

        # valid calls
        mock_connection = Mock()
        mock_connection.server_version = "1.14.0"
        query = GetBuilder("Person", "name", mock_connection).with_near_vector(near_vector).build()
        self.assertEqual(
            "{Get{Person(nearVector: {vector: [1, 2, 3, 4, 5, 6, 7, 8, 9] certainty: 0.55} ){name}}}",
            query,
        )

        # invalid calls
        near_error_msg = (
            "Cannot use multiple 'near' filters, or a 'near' filter along with a 'ask' filter!"
        )

        near_object = {"id": "test_id", "certainty": 0.55}
        with self.assertRaises(AttributeError) as error:
            GetBuilder("Person", "name", mock_connection).with_near_object(
                near_object
            ).with_near_vector(near_vector)
        check_error_message(self, error, near_error_msg)

    def test_build_near_object(self):
        """
        Test the `with_near_object` method.
        """

        near_object = {"id": "test_id", "certainty": 0.55}

        # valid calls
        mock_connection = Mock()
        mock_connection.server_version = "1.14.0"
        query = GetBuilder("Person", "name", mock_connection).with_near_object(near_object).build()
        self.assertEqual('{Get{Person(nearObject: {id: "test_id" certainty: 0.55} ){name}}}', query)

        # invalid calls
        near_error_msg = (
            "Cannot use multiple 'near' filters, or a 'near' filter along with a 'ask' filter!"
        )

        near_text = {
            "concepts": "computer",
            "moveTo": {"concepts": ["science"], "force": 0.5},
        }
        with self.assertRaises(AttributeError) as error:
            GetBuilder("Person", "name", mock_connection).with_near_text(
                near_text
            ).with_near_object(near_object)
        check_error_message(self, error, near_error_msg)

    @patch("weaviate.gql.get.image_encoder_b64", side_effect=lambda x: "test_call")
    def test_build_near_image(self, mock_image_encoder_b64: Mock):
        """
        Test the `with_near_object` method.
        """

        near_image = {"image": "test_image", "certainty": 0.55}

        # valid calls
        ## encode False
        query = GetBuilder("Person", "name", None).with_near_image(near_image, encode=False).build()
        self.assertEqual(
            '{Get{Person(nearImage: {image: "test_image" certainty: 0.55} ){name}}}', query
        )
        mock_image_encoder_b64.assert_not_called()

        ## encode True
        query = GetBuilder("Person", "name", None).with_near_image(near_image, encode=True).build()
        self.assertEqual(
            '{Get{Person(nearImage: {image: "test_call" certainty: 0.55} ){name}}}', query
        )
        mock_image_encoder_b64.assert_called()

        # invalid calls
        near_error_msg = (
            "Cannot use multiple 'near' filters, or a 'near' filter along with a 'ask' filter!"
        )

        near_text = {
            "concepts": "computer",
            "moveTo": {"concepts": ["science"], "force": 0.5},
        }
        with self.assertRaises(AttributeError) as error:
            GetBuilder("Person", "name", None).with_near_text(near_text).with_near_image(near_image)
        check_error_message(self, error, near_error_msg)

    test_build_near_media_param_list: List[
        Tuple[str, str, str, Callable[[GetBuilder, str, str, bool], GetBuilder]]
    ] = [
        (
            "audio",
            "test_audio",
            "nearAudio",
            lambda b, k, v, e: b.with_near_audio({k: v, "certainty": 0.55}, encode=e),
        ),
        (
            "video",
            "test_video",
            "nearVideo",
            lambda b, k, v, e: b.with_near_video({k: v, "certainty": 0.55}, encode=e),
        ),
        (
            "depth",
            "test_depth",
            "nearDepth",
            lambda b, k, v, e: b.with_near_depth({k: v, "certainty": 0.55}, encode=e),
        ),
        (
            "thermal",
            "test_thermal",
            "nearThermal",
            lambda b, k, v, e: b.with_near_thermal({k: v, "certainty": 0.55}, encode=e),
        ),
        (
            "imu",
            "test_imu",
            "nearIMU",
            lambda b, k, v, e: b.with_near_imu({k: v, "certainty": 0.55}, encode=e),
        ),
    ]

    def test_build_near_media(self):
        """
        Test the `with_near_<media>` method.
        """
        for key, value, type_, fn in self.test_build_near_media_param_list:
            with self.subTest(key=key, value=value):
                with patch(
                    "weaviate.gql.get.file_encoder_b64", side_effect=lambda x: "test_call"
                ) as mocked:
                    # valid calls
                    ## encode False
                    query = fn(GetBuilder("Person", "name", None), key, value, False).build()
                    self.assertEqual(
                        f'{{Get{{Person({type_}: {{{key}: "{value}" certainty: 0.55}} ){{name}}}}}}',
                        query,
                    )
                    mocked.assert_not_called()

                    ## encode True
                    query = fn(GetBuilder("Person", "name", None), key, value, True).build()
                    self.assertEqual(
                        f'{{Get{{Person({type_}: {{{key}: "test_call" certainty: 0.55}} ){{name}}}}}}',
                        query,
                    )
                    mocked.assert_called()

                    # invalid calls
                    near_error_msg = "Cannot use multiple 'near' filters, or a 'near' filter along with a 'ask' filter!"

                    near_text = {
                        "concepts": "computer",
                        "moveTo": {"concepts": ["science"], "force": 0.5},
                    }
                    with self.assertRaises(AttributeError) as error:
                        fn(
                            GetBuilder("Person", "name", None).with_near_text(near_text),
                            key,
                            value,
                            True,
                        )
                    check_error_message(self, error, near_error_msg)

    def test_build_ask(self):
        """
        Test the `with_ask` method.
        """

        ask = {
            "question": "What is k8s?",
            "certainty": 0.55,
            "autocorrect": False,
        }

        # valid calls
        query = GetBuilder("Person", "name", None).with_ask(ask).build()
        self.assertEqual(
            '{Get{Person(ask: {question: "What is k8s?" certainty: 0.55 autocorrect: false} ){name}}}',
            query,
        )

        # invalid calls
        near_error_msg = (
            "Cannot use multiple 'near' filters, or a 'near' filter along with a 'ask' filter!"
        )

        near_text = {
            "concepts": "computer",
            "moveTo": {"concepts": ["science"], "force": 0.5},
        }
        with self.assertRaises(AttributeError) as error:
            GetBuilder("Person", "name", None).with_near_text(near_text).with_ask(ask)
        check_error_message(self, error, near_error_msg)

    def test_build_with_additional(self):
        """
        Test the `with_additional` method.
        """

        # valid calls
        ## `str` as argument
        query = GetBuilder("Person", "name", None).with_additional("id").build()
        self.assertEqual("{Get{Person{name _additional {id }}}}", query)

        ## list of `str` as argument
        query = (
            GetBuilder("Person", "name", None).with_additional(["id", "certainty", "test"]).build()
        )
        self.assertEqual("{Get{Person{name _additional {certainty id test }}}}", query)

        ## dict with value `str` as argument
        query = GetBuilder("Person", "name", None).with_additional({"classification": "id"}).build()
        self.assertEqual("{Get{Person{name _additional {classification {id } }}}}", query)

        ## dict with value list of `str` as argument
        query = (
            GetBuilder("Person", "name", None)
            .with_additional({"classification": ["basedOn", "classifiedFields", "completed", "id"]})
            .build()
        )
        self.assertEqual(
            "{Get{Person{name _additional {classification {basedOn classifiedFields completed id } }}}}",
            query,
        )

        ## dict with value list of `tuple` as argument
        clause = {"token": ["entity", "word"]}
        settings = {"test1": 1, "test3": [True], "test2": 10.0}
        query = GetBuilder("Person", "name", None).with_additional((clause, settings)).build()
        self.assertEqual(
            "{Get{Person{name _additional {token(test1: 1 test2: 10.0 test3: [true] ) {entity word } }}}}",
            query,
        )

        ## dict with value list of `tuple` as argument
        clause = {"token": "certainty"}
        settings = {"test1": ["TEST"]}
        query = GetBuilder("Person", "name", None).with_additional((clause, settings)).build()
        self.assertEqual(
            '{Get{Person{name _additional {token(test1: ["TEST"] ) {certainty } }}}}', query
        )

        ## multiple calls
        clause = {"token": "certainty"}
        settings = {"test1": ["TEST"]}
        query = (
            GetBuilder("Person", "name", None)
            .with_additional("test")
            .with_additional(["id", "certainty"])
            .with_additional({"classification": ["completed", "id"]})
            .with_additional((clause, settings))
            .build()
        )
        self.assertEqual(
            '{Get{Person{name _additional {certainty id test classification {completed id } token(test1: ["TEST"] ) {certainty } }}}}',
            query,
        )

        ## multiple calls
        query = (
            GetBuilder("Person", None, None)
            .with_additional("test")
            .with_additional(["id", "certainty"])
            .with_additional({"classification": ["completed", "id"]})
            .with_additional("id")
            .with_additional("test")
            .build()
        )
        self.assertEqual(
            "{Get{Person{ _additional {certainty id test classification {completed id } }}}}", query
        )

        # invalid calls
        # error messages
        prop_type_msg = lambda dt: (
            "The 'properties' argument must be either of type `str`, `list`, `dict` or `tuple`! "
            f"Given: {dt}"
        )
        prop_list_msg = "If type of 'properties' is `list` then all items must be of type `str`!"
        prop_dict_key_msg = "If type of 'properties' is `dict` then all keys must be of type `str`!"
        prop_dict_value_msg = lambda dt: (
            "If type of 'properties' is `dict` then all the values must be either of type "
            f"`str` or `list` of `str`! Given: {dt}!"
        )
        prop_dict_value_len = (
            "If type of 'properties' is `dict` and a value is of type `list` then at least"
            " one element should be present!"
        )
        prop_dict_value_item_msg = (
            "If type of 'properties' is `dict` and a value is of type `list` then all "
            "items must be of type `str`!"
        )
        prop_tuple_len_msg = (
            "If type of 'properties' is `tuple` then it should have length 2: "
            "(clause: <dict>, settings: <dict>)"
        )
        prop_tuple_type_msg = (
            "If type of 'properties' is `tuple` then it should have this data type: "
            "(<dict>, <dict>)"
        )
        prop_tuple_clause_len_msg = lambda clause: (
            "If type of 'properties' is `tuple` then the 'clause' (first element) should "
            f"have only one key. Given: {len(clause)}"
        )
        prop_tuple_settings_len_msg = lambda settings: (
            "If type of 'properties' is `tuple` then the 'settings' (second element) should "
            f"have at least one key. Given: {len(settings)}"
        )
        prop_tuple_clause_key_type_msg = (
            "If type of 'properties' is `tuple` then first element's key should be of type "
            "`str`!"
        )
        prop_tuple_settings_keys_type_msg = (
            "If type of 'properties' is `tuple` then the second elements (<dict>) should "
            "have all the keys of type `str`!"
        )
        prop_tuple_clause_value_type_msg = lambda dt: (
            "If type of 'properties' is `tuple` then first element's dict values must be "
            f"either of type `str` or `list` of `str`! Given: {dt}!"
        )
        prop_tuple_clause_value_len_msg = (
            "If type of 'properties' is `tuple` and first element's dict value is of type "
            "`list` then at least one element should be present!"
        )
        prop_tuple_clause_values_items_type_msg = (
            "If type of 'properties' is `tuple` and first element's dict value is of type "
            " `list` then all items must be of type `str`!"
        )

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional(123)
        check_error_message(self, error, prop_type_msg(int))

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional([123])
        check_error_message(self, error, prop_list_msg)

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional({123: "Test"})
        check_error_message(self, error, prop_dict_key_msg)

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional({"test": True})
        check_error_message(self, error, prop_dict_value_msg(bool))

        with self.assertRaises(ValueError) as error:
            GetBuilder("Person", "name", None).with_additional({"test": []})
        check_error_message(self, error, prop_dict_value_len)

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional({"test": [True]})
        check_error_message(self, error, prop_dict_value_item_msg)

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional({"test": [True]})
        check_error_message(self, error, prop_dict_value_item_msg)

        with self.assertRaises(ValueError) as error:
            GetBuilder("Person", "name", None).with_additional((1,))
        check_error_message(self, error, prop_tuple_len_msg)

        with self.assertRaises(ValueError) as error:
            GetBuilder("Person", "name", None).with_additional((1, 2, 3))
        check_error_message(self, error, prop_tuple_len_msg)

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional(({1: "1"}, ["test"]))
        check_error_message(self, error, prop_tuple_type_msg)

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional(([{1: "1"}], ["test"]))
        check_error_message(self, error, prop_tuple_type_msg)

        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional((["test"], {1: "1"}))
        check_error_message(self, error, prop_tuple_type_msg)

        clause = {"test1": 1, "test2": 2}
        with self.assertRaises(ValueError) as error:
            GetBuilder("Person", "name", None).with_additional((clause, {"1": "1"}))
        check_error_message(self, error, prop_tuple_clause_len_msg(clause))

        clause = {"test1": "1"}
        settings = {}
        with self.assertRaises(ValueError) as error:
            GetBuilder("Person", "name", None).with_additional((clause, settings))
        check_error_message(self, error, prop_tuple_settings_len_msg(settings))

        clause = {1: "1"}
        settings = {"test": 1}
        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional((clause, settings))
        check_error_message(self, error, prop_tuple_clause_key_type_msg)

        clause = {"test": "1"}
        settings = {"test": 1, 2: 2}
        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional((clause, settings))
        check_error_message(self, error, prop_tuple_settings_keys_type_msg)

        clause = {"test": "1"}
        settings = {2: 2}
        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional((clause, settings))
        check_error_message(self, error, prop_tuple_settings_keys_type_msg)

        clause = {"test": True}
        settings = {"test": 2}
        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional((clause, settings))
        check_error_message(self, error, prop_tuple_clause_value_type_msg(bool))

        clause = {"test": []}
        settings = {"test": 2}
        with self.assertRaises(ValueError) as error:
            GetBuilder("Person", "name", None).with_additional((clause, settings))
        check_error_message(self, error, prop_tuple_clause_value_len_msg)

        clause = {"test": ["1", "2", 3]}
        settings = {"test": 2}
        with self.assertRaises(TypeError) as error:
            GetBuilder("Person", "name", None).with_additional((clause, settings))
        check_error_message(self, error, prop_tuple_clause_values_items_type_msg)

    def test_build(self):
        """
        Test the `build` method. (without filters)
        """

        error_message = (
            "No 'properties' or 'additional properties' specified to be returned. "
            "At least one should be included."
        )

        with self.assertRaises(AttributeError) as error:
            query = GetBuilder("Group", [], None).build()
        check_error_message(self, error, error_message)

        query = GetBuilder("Group", "name", None).build()
        self.assertEqual("{Get{Group{name}}}", query)

        query = GetBuilder("Group", ["name", "uuid"], None).build()
        self.assertEqual("{Get{Group{name uuid}}}", query)

        query = GetBuilder("Group", None, None).with_additional("distance").build()
        self.assertEqual("{Get{Group{ _additional {distance }}}}", query)

        near_text = {
            "concepts": ["computer"],
            "moveTo": {"concepts": "science", "force": 0.1},
            "moveAwayFrom": {"concepts": ["airplane"], "force": 0.2},
            "certainty": 0.3,
        }
        or_filter = {
            "operator": "Or",
            "operands": [
                {
                    "path": ["name"],
                    "operator": "Equal",
                    "valueString": "Alan Turing",
                },
                {"path": ["name"], "operator": "Equal", "valueString": "John von Neumann"},
            ],
        }
        query = (
            GetBuilder("Person", ["name", "uuid"], None)
            .with_near_text(near_text)
            .with_where(or_filter)
            .with_limit(2)
            .with_offset(10)
            .build()
        )
        self.assertEqual(
            '{Get{Person(where: {operator: Or operands: [{path: ["name"] operator: Equal valueString: "Alan Turing"}, {path: ["name"] operator: Equal valueString: "John von Neumann"}]} limit: 2 offset: 10 nearText: {concepts: ["computer"] certainty: 0.3 moveTo: {force: 0.1 concepts: ["science"]} moveAwayFrom: {force: 0.2 concepts: ["airplane"]}} ){name uuid}}}',
            query,
        )

    def test_capitalized_class_name(self):
        """
        Test the capitalized class_name.
        """

        get = GetBuilder("Test", ["prop"], None)
        self.assertEqual(get._class_name, "Test")

        get = GetBuilder("test", ["prop"], None)
        self.assertEqual(get._class_name, "Test")



import unittest
from unittest.mock import patch

from test.util import check_error_message
from weaviate.connect.connection import (
    Connection,
    _get_proxies,
)
from weaviate.util import _get_valid_timeout_config


class TestConnection(unittest.TestCase):
    def check_connection_attributes(
        self,
        connection: Connection,
        url="test_url",
        timeout_config=(2, 20),
        oidc_auth_flow=False,
        headers=None,
    ):
        """
        Check the attributes of the connection value. Assign 'skip' to
        an attribute to skip testing. The attributes have the default constructor values.
        """

        if headers is None:
            headers = {"content-type": "application/json"}
        if url != "skip":
            self.assertEqual(connection.url, url)
        if timeout_config != "skip":
            self.assertEqual(connection.timeout_config, timeout_config)
        if oidc_auth_flow != "skip":
            if oidc_auth_flow is True:
                self.assertIsNotNone(connection._auth)
            else:
                self.assertIsNone(connection._auth)
        if headers != "skip":
            self.assertEqual(connection._headers, headers)

    @patch("weaviate.connect.connection.datetime")
    def test_get_epoch_time(self, mock_datetime):
        """
        Test the `get_epoch_time` function.
        """

        import datetime
        from weaviate.connect.connection import _get_epoch_time

        zero_epoch = datetime.datetime.fromtimestamp(0)
        mock_datetime.datetime.utcnow.return_value = zero_epoch
        self.assertEqual(_get_epoch_time(), 0)

        epoch = datetime.datetime.fromtimestamp(110.56)
        mock_datetime.datetime.utcnow.return_value = epoch
        self.assertEqual(_get_epoch_time(), 111)

        epoch = datetime.datetime.fromtimestamp(110.46)
        mock_datetime.datetime.utcnow.return_value = epoch
        self.assertEqual(_get_epoch_time(), 110)

    @patch("weaviate.connect.connection.os")
    def test_get_proxies(self, os_mock):
        """
        Test the `_get_proxies` function.
        """

        error_msg = lambda dt: (
            "If 'proxies' is not None, it must be of type dict or str. " f"Given type: {dt}."
        )
        with self.assertRaises(TypeError) as error:
            proxies = _get_proxies([], False)
        check_error_message(self, error, error_msg(list))

        proxies = _get_proxies({}, False)
        self.assertEqual(proxies, {})

        proxies = _get_proxies({"test": True}, False)
        self.assertEqual(proxies, {"test": True})

        proxies = _get_proxies({"test": True}, True)
        self.assertEqual(proxies, {"test": True})

        proxies = _get_proxies("test", True)
        self.assertEqual(proxies, {"http": "test", "https": "test"})

        os_mock.environ.get.return_value = None
        proxies = _get_proxies(None, True)
        self.assertEqual(proxies, {})

        os_mock.environ.get.return_value = "test"
        proxies = _get_proxies(None, True)
        self.assertEqual(proxies, {"http": "test", "https": "test"})

    def test__get_valid_timeout_config(self):
        """
        Test the `_get_valid_timeout_config` function.
        """

        # incalid calls
        negative_num_error_message = "'timeout_config' cannot be non-positive number/s!"
        type_error_message = "'timeout_config' should be a (or tuple of) positive number/s!"
        value_error_message = "'timeout_config' must be of length 2!"
        value_types_error_message = "'timeout_config' must be tuple of numbers"

        ## wrong type
        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config(None)
        check_error_message(self, error, type_error_message)

        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config(True)
        check_error_message(self, error, type_error_message)

        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config("(2, 13)")
        check_error_message(self, error, type_error_message)

        ## wrong tuple length 3
        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config((1, 2, 3))
        check_error_message(self, error, value_error_message)

        ## wrong value types
        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config((None, None))
        check_error_message(self, error, value_types_error_message)

        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config(("1", 10))
        check_error_message(self, error, value_types_error_message)

        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config(("1", "10"))
        check_error_message(self, error, value_types_error_message)

        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config((True, False))
        check_error_message(self, error, value_types_error_message)

        ## non-positive numbers
        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config(0)
        check_error_message(self, error, negative_num_error_message)

        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config(-1)
        check_error_message(self, error, negative_num_error_message)

        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config(-4.134)
        check_error_message(self, error, negative_num_error_message)

        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config((-3.5, 1.5))
        check_error_message(self, error, negative_num_error_message)

        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config((3, -1.5))
        check_error_message(self, error, negative_num_error_message)

        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config((0, 0))
        check_error_message(self, error, negative_num_error_message)

        # valid calls
        self.assertEqual(_get_valid_timeout_config((2, 20)), (2, 20))
        self.assertEqual(_get_valid_timeout_config((3.5, 2.34)), (3.5, 2.34))
        self.assertEqual(_get_valid_timeout_config(4.32), (4.32, 4.32))



import pytest
from pydantic import ValidationError

from weaviate.collections.classes.grpc import FromReference, FromReferenceMultiTarget


def test_link_to_errors_on_extra_variable():
    with pytest.raises(ValidationError):
        FromReference(link_on="ref", return_property="name")


def test_link_to_multi_target_errors_on_extra_variable():
    with pytest.raises(ValidationError):
        FromReferenceMultiTarget(link_on="ref", target_colection="Test")


def test_link_to_multi_target_is_link_to():
    link_to = FromReferenceMultiTarget(link_on="ref", target_collection="Test")
    assert isinstance(link_to, FromReference)



import json
from typing import Optional

import pytest
import uuid
from werkzeug.wrappers import Request, Response

import weaviate
from mock_tests.conftest import MOCK_SERVER_URL
from weaviate.batch.crud_batch import WeaviateErrorRetryConf, BatchResponse
from weaviate.util import check_batch_result


@pytest.mark.parametrize(
    "error",
    [
        {"errors": {"error": [{"message": "I'm an error message"}]}},
        {
            "errors": {
                "error": [{"message": "I'm an error message"}, {"message": "Another message"}]
            }
        },
    ],
)
def test_automatic_retry_obs(weaviate_mock, error):
    """Tests that all objects are successfully added even if half of them fail."""
    successfully_added = []
    num_failed_requests = 0

    # Mockserver returns error for half of all objects
    def handler(request: Request):
        nonlocal num_failed_requests
        objects = request.json["objects"]
        for j, obj in enumerate(objects):
            obj["deprecations"] = None
            if j % 2 == 0:
                obj["result"] = {}
                successfully_added.append(uuid.UUID(obj["id"]))
            else:
                obj["result"] = error
                num_failed_requests += 1
        return Response(json.dumps(objects))

    weaviate_mock.expect_request("/v1/batch/objects").respond_with_handler(handler)

    client = weaviate.Client(MOCK_SERVER_URL)
    added_uuids = []
    batch_size = 4  # Do not change, affects how many failed requests there are
    n = (
        50 * batch_size
    )  # multiple of the batch size, otherwise it is difficult to calculate the number of expected errors
    client.batch.configure(
        batch_size=batch_size,
        num_workers=2,
        weaviate_error_retries=WeaviateErrorRetryConf(number_retries=3),
        dynamic=False,
    )

    with client.batch as batch:
        for i in range(n):
            added_uuids.append(uuid.uuid4())
            batch.add_data_object({"name": "test" + str(i)}, "test", added_uuids[-1])
    assert len(successfully_added) == n
    assert sorted(successfully_added) == sorted(added_uuids)

    # with a batch size of 4, we have 3 failures per batch
    assert num_failed_requests == 3 * n / batch_size


def test_automatic_retry_refs(weaviate_mock):
    """Tests that all references are successfully added even if half of them fail."""
    num_success_requests = 0
    num_failed_requests = 0

    # Mockserver returns error for half of all objects
    def handler(request: Request):
        nonlocal num_failed_requests, num_success_requests
        objects = request.json
        for j, obj in enumerate(objects):
            obj["deprecations"] = None
            if j % 2 == 0:
                obj["result"] = {}
                num_success_requests += 1
            else:
                obj["result"] = {"errors": {"error": [{"message": "I'm an error message"}]}}
                num_failed_requests += 1
        return Response(json.dumps(objects))

    weaviate_mock.expect_request("/v1/batch/references").respond_with_handler(handler)

    client = weaviate.Client(MOCK_SERVER_URL)
    batch_size = 4  # Do not change, affects how many failed requests there are
    n = (
        50 * batch_size
    )  # multiple of the batch size, otherwise it is difficult to calculate the number of expected errors
    with client.batch(
        batch_size=batch_size,
        weaviate_error_retries=WeaviateErrorRetryConf(number_retries=3),
        num_workers=2,
        dynamic=False,
    ) as batch:
        for _ in range(n):
            batch.add_reference(
                from_property_name="Property",
                from_object_class_name="SomeClass",
                from_object_uuid=str(uuid.uuid4()),
                to_object_class_name="otherClass",
                to_object_uuid=str(uuid.uuid4()),
            )
    assert num_success_requests == n

    # with a batch size of 4, we have 3 failures per batch
    assert num_failed_requests == 3 * n / batch_size


def test_automatic_retry_unsuccessful(weaviate_mock):
    """Test automatic retry that cannot add all objects."""
    num_success_requests = 0

    # Mockserver returns error for half of all objects
    def handler(request: Request):
        nonlocal num_success_requests
        objects = request.json["objects"]
        for j, obj in enumerate(objects):
            obj["deprecations"] = None
            if j % 2 == 0:
                obj["result"] = {}
                num_success_requests += 1
            else:
                obj["result"] = {"errors": {"error": [{"message": "I'm an error message"}]}}
        return Response(json.dumps(objects))

    weaviate_mock.expect_request("/v1/batch/objects").respond_with_handler(handler)

    client = weaviate.Client(MOCK_SERVER_URL)
    batch_size = 20
    n = batch_size * 2
    with client.batch(
        batch_size=batch_size,
        weaviate_error_retries=WeaviateErrorRetryConf(number_retries=1),
        num_workers=2,
        callback=None,
    ) as batch:
        for i in range(n):
            batch.add_data_object({"name": "test" + str(i)}, "test", uuid.uuid4())
        batch.flush()
    # retried 3 times, starting with 200 objects and half off all objects succeed each time
    assert num_success_requests == 30


@pytest.mark.parametrize(
    "retry_config",
    [None, WeaviateErrorRetryConf(number_retries=1), WeaviateErrorRetryConf(number_retries=2)],
)
def test_print_threadsafety(weaviate_mock, capfd, retry_config):
    """Test retry with callback and callback threadsafety."""
    num_success_requests = 0

    # Mockserver returns error for half of all objects
    def handler(request: Request):
        nonlocal num_success_requests
        objects = request.json["objects"]
        for j, obj in enumerate(objects):
            obj["deprecations"] = None
            if j % 2 == 0:
                obj["result"] = {}
                num_success_requests += 1
            else:
                obj["result"] = {"errors": {"error": [{"message": "I'm an error message"}]}}
        return Response(json.dumps(objects))

    weaviate_mock.expect_request("/v1/batch/objects").respond_with_handler(handler)

    client = weaviate.Client(MOCK_SERVER_URL)

    added_uuids = []
    n = 200 * 4
    with client.batch(
        batch_size=200,
        callback=check_batch_result,
        num_workers=4,
        weaviate_error_retries=retry_config,
    ) as batch:
        for i in range(n):
            added_uuids.append(uuid.uuid4())
            batch.add_data_object({"name": "test" + str(i)}, "test", added_uuids[-1])

    retry_factor: float = 1.0
    if retry_config is not None:
        retry_factor = 1 / (2 * retry_config.number_retries)
    assert num_success_requests == n - n / 2 * retry_factor

    # half of all objects fail => N/2 print statements that end with a newline
    print_output, err = capfd.readouterr()
    assert print_output.count("\n") == n - num_success_requests


@pytest.mark.parametrize(
    "retry_config, expected",
    [
        (WeaviateErrorRetryConf(number_retries=1, errors_to_include=["include", "maybe"]), 300),
        (WeaviateErrorRetryConf(number_retries=1, errors_to_exclude=["reject", "maybe"]), 250),
    ],
)
def test_include_error(weaviate_mock, retry_config, expected):
    """Test that objects are included/excluded based on their error message"""
    num_success_requests = 0

    # Mockserver returns error for 3/4 of all objects, with different messages for each quarter
    def handler(request: Request):
        nonlocal num_success_requests
        objects = request.json["objects"]
        for j, obj in enumerate(objects):
            obj["deprecations"] = None
            if j % 4 == 0:
                obj["result"] = {}
                num_success_requests += 1
            elif j % 4 == 1:
                obj["result"] = {"errors": {"error": [{"message": "include me"}]}}
            elif j % 4 == 2:
                obj["result"] = {"errors": {"error": [{"message": "maybe retry maybe not"}]}}
            else:
                obj["result"] = {
                    "errors": {"error": [{"message": "reject me"}, {"message": "other error"}]}
                }
        return Response(json.dumps(objects))

    weaviate_mock.expect_request("/v1/batch/objects").respond_with_handler(handler)

    client = weaviate.Client(MOCK_SERVER_URL)

    added_uuids = []
    n = 400 * 2
    with client.batch(
        batch_size=400,
        callback=None,
        num_workers=2,
        weaviate_error_retries=retry_config,
    ) as batch:
        for i in range(n):
            added_uuids.append(uuid.uuid4())
            batch.add_data_object({"name": "test" + str(i)}, "test", added_uuids[-1])

    assert num_success_requests == expected


def test_callback_for_successful_responses(weaviate_mock, capfd):
    """Test that all objects reach teh callback, even when a part of a batch is retried."""

    # have some objects fail
    def handler(request: Request):
        objects = request.json["objects"]
        for j, obj in enumerate(objects):
            obj["deprecations"] = None
            if j % 2 == 0:
                obj["result"] = {}
            else:
                obj["result"] = {"errors": {"error": [{"message": "I'm an error message"}]}}
        return Response(json.dumps(objects))

    weaviate_mock.expect_request("/v1/batch/objects").respond_with_handler(handler)

    client = weaviate.Client(MOCK_SERVER_URL)

    def callback_print_all(results: Optional[BatchResponse]):
        if results is None:
            return
        for _ in results:
            print("I saw that object")

    added_uuids = []
    n = 200 * 4
    with client.batch(
        batch_size=200,
        callback=callback_print_all,
        num_workers=4,
        weaviate_error_retries=WeaviateErrorRetryConf(number_retries=1),
    ) as batch:
        for i in range(n):
            added_uuids.append(uuid.uuid4())
            batch.add_data_object({"name": "test" + str(i)}, "test", added_uuids[-1])

    # callback output for each object
    print_output, err = capfd.readouterr()
    assert print_output.count("\n") == n



from typing import List, TypedDict, Union

import pytest

import weaviate
from weaviate.collections.classes.config import (
    DataType,
    Property,
)
from weaviate.collections.classes.grpc import PROPERTIES, FromNested
from weaviate.collections.classes.internal import Nested


@pytest.fixture(scope="module")
def client():
    client = weaviate.connect_to_local()
    client.collections.delete_all()
    yield client
    client.collections.delete_all()


@pytest.mark.parametrize(
    "property_,object_",
    [
        (
            Property(
                name="nested",
                data_type=DataType.OBJECT,
                nested_properties=Property(
                    name="text",
                    data_type=DataType.TEXT,
                ),
            ),
            {"text": "Hello World"},
        ),
        (
            Property(
                name="nested",
                data_type=DataType.OBJECT,
                nested_properties=[
                    Property(
                        name="text",
                        data_type=DataType.TEXT,
                    )
                ],
            ),
            {"text": "Hello World"},
        ),
        (
            Property(
                name="nested",
                data_type=DataType.OBJECT_ARRAY,
                nested_properties=Property(
                    name="text",
                    data_type=DataType.TEXT,
                ),
            ),
            [{"text": "Hello World"}, {"text": "Hello World"}],
        ),
        (
            Property(
                name="nested",
                data_type=DataType.OBJECT_ARRAY,
                nested_properties=[
                    Property(
                        name="text",
                        data_type=DataType.TEXT,
                    )
                ],
            ),
            [{"text": "Hello World"}, {"text": "Hello World"}],
        ),
        (
            Property(
                name="nested",
                data_type=DataType.OBJECT,
                nested_properties=[
                    Property(
                        name="text",
                        data_type=DataType.TEXT,
                    ),
                    Property(
                        name="texts",
                        data_type=DataType.TEXT_ARRAY,
                    ),
                    Property(
                        name="number",
                        data_type=DataType.NUMBER,
                    ),
                    Property(
                        name="numbers",
                        data_type=DataType.NUMBER_ARRAY,
                    ),
                    Property(
                        name="int",
                        data_type=DataType.INT,
                    ),
                    Property(
                        name="ints",
                        data_type=DataType.INT_ARRAY,
                    ),
                    Property(
                        name="bool",
                        data_type=DataType.BOOL,
                    ),
                    Property(
                        name="bools",
                        data_type=DataType.BOOL_ARRAY,
                    ),
                    Property(
                        name="date",
                        data_type=DataType.DATE,
                    ),
                    Property(
                        name="dates",
                        data_type=DataType.DATE_ARRAY,
                    ),
                    Property(
                        name="obj",
                        data_type=DataType.OBJECT,
                        nested_properties=Property(
                            name="text",
                            data_type=DataType.TEXT,
                        ),
                    ),
                    Property(
                        name="objs",
                        data_type=DataType.OBJECT_ARRAY,
                        nested_properties=Property(
                            name="text",
                            data_type=DataType.TEXT,
                        ),
                    ),
                ],
            ),
            {
                "text": "Hello World",
                "texts": ["Hello", "World"],
                "number": 42.0,
                "numbers": [42.0, 43.0],
                "int": 42,
                "ints": [42, 43],
                "bool": True,
                "bools": [True, False],
                "date": "2020-01-01T00:00:00Z",
                "dates": ["2020-01-01T00:00:00Z", "2020-01-02T00:00:00Z"],
                "obj": {"text": "Hello World"},
                "objs": [{"text": "Hello World"}, {"text": "Hello World"}],
            },
        ),
        (
            Property(
                name="nested",
                data_type=DataType.OBJECT_ARRAY,
                nested_properties=[
                    Property(
                        name="text",
                        data_type=DataType.TEXT,
                    ),
                    Property(
                        name="texts",
                        data_type=DataType.TEXT_ARRAY,
                    ),
                    Property(
                        name="number",
                        data_type=DataType.NUMBER,
                    ),
                    Property(
                        name="numbers",
                        data_type=DataType.NUMBER_ARRAY,
                    ),
                    Property(
                        name="int",
                        data_type=DataType.INT,
                    ),
                    Property(
                        name="ints",
                        data_type=DataType.INT_ARRAY,
                    ),
                    Property(
                        name="bool",
                        data_type=DataType.BOOL,
                    ),
                    Property(
                        name="bools",
                        data_type=DataType.BOOL_ARRAY,
                    ),
                    Property(
                        name="date",
                        data_type=DataType.DATE,
                    ),
                    Property(
                        name="dates",
                        data_type=DataType.DATE_ARRAY,
                    ),
                    Property(
                        name="obj",
                        data_type=DataType.OBJECT,
                        nested_properties=Property(
                            name="text",
                            data_type=DataType.TEXT,
                        ),
                    ),
                    Property(
                        name="objs",
                        data_type=DataType.OBJECT_ARRAY,
                        nested_properties=Property(
                            name="text",
                            data_type=DataType.TEXT,
                        ),
                    ),
                ],
            ),
            [
                {
                    "text": "Hello World",
                    "texts": ["Hello", "World"],
                    "number": 42.0,
                    "numbers": [42.0, 43.0],
                    "int": 42,
                    "ints": [42, 43],
                    "bool": True,
                    "bools": [True, False],
                    "date": "2020-01-01T00:00:00Z",
                    "dates": ["2020-01-01T00:00:00Z", "2020-01-02T00:00:00Z"],
                    "obj": {"text": "Hello World"},
                    "objs": [{"text": "Hello World"}, {"text": "Hello World"}],
                }
            ],
        ),
        (
            Property(
                name="nested",
                data_type=DataType.OBJECT,
                nested_properties=Property(
                    name="a",
                    data_type=DataType.OBJECT,
                    nested_properties=Property(
                        name="b",
                        data_type=DataType.OBJECT,
                        nested_properties=Property(
                            name="c",
                            data_type=DataType.OBJECT,
                            nested_properties=Property(
                                name="d",
                                data_type=DataType.TEXT,
                            ),
                        ),
                    ),
                ),
            ),
            {"a": {"b": {"c": {"d": "e"}}}},
        ),
    ],
)
def test_nested_return_all_properties(
    client: weaviate.Client, property_: Property, object_: Union[dict, List[dict]]
):
    name = "TestInsertNestedPropertiesAll"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[property_],
    )
    res = collection.data.insert_many([{"nested": object_}])
    assert res.has_errors is False

    result = collection.query.fetch_objects()
    assert result.objects[0].properties["nested"] == object_


@pytest.mark.parametrize(
    "return_properties,expected",
    [
        (FromNested(name="nested", properties=["text"]), {"text": "Hello World"}),
        (FromNested(name="nested", properties=["texts"]), {"texts": ["Hello", "World"]}),
        (FromNested(name="nested", properties=["number"]), {"number": 42.0}),
        (FromNested(name="nested", properties=["numbers"]), {"numbers": [42.0, 43.0]}),
        (FromNested(name="nested", properties=["int"]), {"int": 42}),
        (FromNested(name="nested", properties=["ints"]), {"ints": [42, 43]}),
        (FromNested(name="nested", properties=["bool"]), {"bool": True}),
        (FromNested(name="nested", properties=["bools"]), {"bools": [True, False]}),
        (FromNested(name="nested", properties=["date"]), {"date": "2020-01-01T00:00:00Z"}),
        (
            FromNested(name="nested", properties=["dates"]),
            {"dates": ["2020-01-01T00:00:00Z", "2020-01-02T00:00:00Z"]},
        ),
        (
            FromNested(name="nested", properties=[FromNested(name="obj", properties=["text"])]),
            {"obj": {"text": "Hello World"}},
        ),
        (
            FromNested(name="nested", properties=[FromNested(name="objs", properties=["text"])]),
            {"objs": [{"text": "Hello World"}, {"text": "Hello World"}]},
        ),
        (
            FromNested(
                name="nested",
                properties=FromNested(
                    name="a",
                    properties=[
                        FromNested(name="b", properties=[FromNested(name="c", properties=["d"])])
                    ],
                ),
            ),
            {"a": {"b": {"c": {"d": "e"}}}},
        ),
    ],
)
def test_nested_return_specific_properties(
    client: weaviate.Client, return_properties: PROPERTIES, expected: dict
):
    name = "TestInsertNestedPropertiesSpecific"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[
            Property(
                name="nested",
                data_type=DataType.OBJECT,
                nested_properties=[
                    Property(
                        name="text",
                        data_type=DataType.TEXT,
                    ),
                    Property(
                        name="texts",
                        data_type=DataType.TEXT_ARRAY,
                    ),
                    Property(
                        name="number",
                        data_type=DataType.NUMBER,
                    ),
                    Property(
                        name="numbers",
                        data_type=DataType.NUMBER_ARRAY,
                    ),
                    Property(
                        name="int",
                        data_type=DataType.INT,
                    ),
                    Property(
                        name="ints",
                        data_type=DataType.INT_ARRAY,
                    ),
                    Property(
                        name="bool",
                        data_type=DataType.BOOL,
                    ),
                    Property(
                        name="bools",
                        data_type=DataType.BOOL_ARRAY,
                    ),
                    Property(
                        name="date",
                        data_type=DataType.DATE,
                    ),
                    Property(
                        name="dates",
                        data_type=DataType.DATE_ARRAY,
                    ),
                    Property(
                        name="obj",
                        data_type=DataType.OBJECT,
                        nested_properties=Property(
                            name="text",
                            data_type=DataType.TEXT,
                        ),
                    ),
                    Property(
                        name="objs",
                        data_type=DataType.OBJECT_ARRAY,
                        nested_properties=Property(
                            name="text",
                            data_type=DataType.TEXT,
                        ),
                    ),
                    Property(
                        name="a",
                        data_type=DataType.OBJECT,
                        nested_properties=Property(
                            name="b",
                            data_type=DataType.OBJECT,
                            nested_properties=Property(
                                name="c",
                                data_type=DataType.OBJECT,
                                nested_properties=Property(
                                    name="d",
                                    data_type=DataType.TEXT,
                                ),
                            ),
                        ),
                    ),
                ],
            ),
        ],
    )
    res = collection.data.insert_many(
        [
            {
                "nested": {
                    "text": "Hello World",
                    "texts": ["Hello", "World"],
                    "number": 42.0,
                    "numbers": [42.0, 43.0],
                    "int": 42,
                    "ints": [42, 43],
                    "bool": True,
                    "bools": [True, False],
                    "date": "2020-01-01T00:00:00Z",
                    "dates": ["2020-01-01T00:00:00Z", "2020-01-02T00:00:00Z"],
                    "obj": {"text": "Hello World"},
                    "objs": [{"text": "Hello World"}, {"text": "Hello World"}],
                    "a": {"b": {"c": {"d": "e"}}},
                }
            }
        ]
    )
    assert res.has_errors is False
    result = collection.query.fetch_objects(return_properties=return_properties)
    assert result.objects[0].properties["nested"] == expected


def test_nested_return_generic_properties(
    client: weaviate.Client,
):
    name = "TestInsertNestedPropertiesGeneric"
    client.collections.delete(name)

    class Child(TypedDict):
        name: str
        age: int

    class Parent(TypedDict):
        child: Nested[Child]

    collection = client.collections.create(
        name=name,
        properties=[
            Property(
                name="child",
                data_type=DataType.OBJECT,
                nested_properties=[
                    Property(
                        name="name",
                        data_type=DataType.TEXT,
                    ),
                    Property(
                        name="age",
                        data_type=DataType.INT,
                    ),
                ],
            )
        ],
        data_model=Parent,
    )

    collection.data.insert(Parent(child=Child(name="Timmy", age=10)))
    results = collection.query.fetch_objects(return_properties=Parent)
    assert results.objects[0].properties["child"] == {"name": "Timmy", "age": 10}
